"""
Creates objective functions by providing a common interface to all factories in the repository.
"""

import configparser
from pathlib import Path

from poli.core import registry
from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.problem import Problem
from poli.core.registry import (
    _DEFAULT,
    _DEFAULT_OBSERVER_RUN_SCRIPT,
    _OBSERVER,
    DEFAULT_OBSERVER_NAME,
)
from poli.core.util.abstract_observer import AbstractObserver
from poli.core.util.algorithm_observer_wrapper import AlgorithmObserverWrapper
from poli.core.util.default_observer import DefaultObserver
from poli.core.util.external_observer import ExternalObserver
from poli.external_isolated_function_script import dynamically_instantiate
from poli.objective_repository import AVAILABLE_OBJECTIVES, AVAILABLE_PROBLEM_FACTORIES


def load_config():
    """Loads the configuration file containing which objectives are registered.

    Returns
    -------
    config : configparser.ConfigParser
        The configuration file.

    """
    HOME_DIR = Path.home().resolve()
    config_file = str(HOME_DIR / ".poli_objectives" / "config.rc")
    config = configparser.ConfigParser(defaults={_OBSERVER: ""})
    _ = config.read(config_file)

    return config


def __create_problem_from_repository(
    name: str,
    seed: int = None,
    batch_size: int = None,
    parallelize: bool = False,
    num_workers: int = None,
    evaluation_budget: int = float("inf"),
    force_isolation: bool = False,
    observer: AbstractObserver = None,
    **kwargs_for_factory,
) -> Problem:
    """Creates the objective function from the repository.

    We create the problem directly, without starting it as an isolated process.

    Parameters
    ----------
    name : str
        The name of the objective function.
    seed : int, optional
        The seed value for random number generation.
    batch_size : int, optional
        The batch size, passed to the black box to run evaluations on batches.
        If None, it will evaluate all inputs at once.
    parallelize : bool, optional
        If True, then the objective function runs in parallel.
    num_workers : int, optional
        When parallelize is True, this specifies the number of processes to use.
        By default, it uses half the number of available CPUs (rounded down).
    evaluation_budget : int, optional
        The maximum number of evaluations allowed. By default, it is infinity.
    observer : AbstractObserver, optional
        The observer to use.
    **kwargs_for_factory : dict, optional
        Additional keyword arguments for the factory.
    """
    if name not in AVAILABLE_PROBLEM_FACTORIES:
        if name in AVAILABLE_OBJECTIVES:
            # For poli developers:
            # If you get this error while developing a new black box,
            # it might be because you forgot to register your problem
            # factory and black box inside the __init__.py file of
            # the objective_repository. Remember that you need to add
            # them to AVAILABLE_PROBLEM_FACTORIES and AVAILABLE_BLACK_BOXES.
            raise ValueError(
                f"Objective function '{name}' is available in the repository, "
                "but it is not registered as available. This is an internal error. "
                "We encourage you to report it to the developers by creating an "
                "issue in the GitHub repository of poli."
            )
        else:
            raise ValueError(
                f"Objective function '{name}' is not available in the repository."
            )
    problem_factory: AbstractProblemFactory = AVAILABLE_PROBLEM_FACTORIES[name]()
    problem = problem_factory.create(
        seed=seed,
        batch_size=batch_size,
        parallelize=parallelize,
        num_workers=num_workers,
        evaluation_budget=evaluation_budget,
        force_isolation=force_isolation,
        **kwargs_for_factory,
    )

    if observer is not None:
        problem.black_box.set_observer(observer)

    return problem


def create(
    name: str,
    *,
    seed: int = None,
    observer_init_info: dict = None,
    observer_name: str = None,
    force_isolation: bool = False,
    batch_size: int = None,
    parallelize: bool = False,
    num_workers: int = None,
    evaluation_budget: int = float("inf"),
    quiet: bool = False,
    **kwargs_for_factory,
) -> Problem:
    """
    Instantiantes a black-box function by calling the `create` method of the associated factory.

    Parameters
    ----------
    name : str
        The name of the objective function.
    seed : int, optional
        The seed value for random number generation.
    observer_init_info : dict, optional
        Optional information about the caller that is forwarded to the logger to initialize the run.
    observer_name : str, optional
        The observer to use.
    force_isolation : bool, optional
        If True, then the objective function is instantiated as an isolated
        process.
    batch_size : int, optional
        The batch size, passed to the black box to run evaluations on batches.
        If None, it will evaluate all inputs at once.
    parallelize : bool, optional
        If True, then the objective function runs in parallel.
    num_workers : int, optional
        When parallelize is True, this specifies the number of processes to use.
        By default, it uses half the number of available CPUs (rounded down).
    evaluation_budget : int, optional
        The maximum number of evaluations allowed. By default, it is infinity.
    quiet : bool, optional
        If True, we squelch the messages giving feedback about the creation process.
        By default, it is False.
    **kwargs_for_factory : dict, optional
        Additional keyword arguments for the factory.

    Returns
    -------
    problem : AbstractProblem
        The black-box function, initial value, and related information.
    """
    problem = __create_problem_from_repository(
        name,
        seed=seed,
        batch_size=batch_size,
        parallelize=parallelize,
        num_workers=num_workers,
        evaluation_budget=evaluation_budget,
        force_isolation=force_isolation,
        **kwargs_for_factory,
    )

    # instantiate observer (if desired)
    observer = _instantiate_observer(observer_name, quiet)
    observer_info = observer.initialize_observer(
        problem.black_box.info, observer_init_info, seed
    )

    f = problem.black_box
    f.set_observer(observer)
    f.set_observer_info(observer_info)
    problem.set_observer(AlgorithmObserverWrapper(observer), observer_info)

    return problem


def start(
    name: str,
    seed: int = None,
    caller_info: dict = None,
    observer_name: str = None,
    force_isolation: bool = False,
    **kwargs_for_factory,
) -> AbstractBlackBox:
    """Starts the black-box function.

    Works just like create, but it returns only the black-box function
    and resets the evaluation budget.

    One example of running this function:

    ```python
    from poli import objective_factory

    with objective_factory.start(name="aloha") as f:
        x = np.array(["F", "L", "E", "A", "S"]).reshape(1, -1)
        y = f(x)
        print(x, y)
    ```

    If the black-box function is created as an isolated process, then
    it will automatically close when the context manager is closed.

    Parameters
    ----------
    name : str
        The name of the objective function.
    seed : int, optional
        The seed value for random number generation.
    caller_info : dict, optional
        Optional information about the caller that is forwarded to the logger to initialize the run.
    observer_name : str, optional
        The observer to use.
    force_isolation : bool, optional
        If True, then the objective function is instantiated as an isolated
        process.
    **kwargs_for_factory : dict, optional
        Additional keyword arguments for the factory.
    """
    problem = create(
        name=name,
        seed=seed,
        observer_init_info=caller_info,
        observer_name=observer_name,
        force_isolation=force_isolation,
        **kwargs_for_factory,
    )

    f = problem.black_box
    # Reset the counter of evaluations
    f.reset_evaluation_budget()

    return f


def _instantiate_observer(observer_name: str, quiet: bool = False) -> AbstractObserver:
    """
    This function attempts to locally instantiate an observer and if that fails starts the observer in the dedicated environment.

    Parameters
    ----------
    observer_name : str
        The observer to use.
    quiet : bool, optional
        If True, we squelch the messages giving feedback about the creation process.

    Returns
    -------
    observer : AbstractObserver
        The observer, either dynamically instantiated or started as an isolated process.
    """
    if _OBSERVER not in registry.config[_DEFAULT]:
        registry.config[_DEFAULT][_OBSERVER] = _DEFAULT_OBSERVER_RUN_SCRIPT

    observer_script: str = registry.config[_DEFAULT][_OBSERVER]
    if observer_name is not None:
        if observer_name != DEFAULT_OBSERVER_NAME:
            observer_script = registry.config[_OBSERVER][observer_name]
        else:
            observer_script = _DEFAULT_OBSERVER_RUN_SCRIPT

    if observer_script == _DEFAULT_OBSERVER_RUN_SCRIPT:
        observer = DefaultObserver()
    else:
        if not quiet:
            print("poli 🧪: initializing the observer.")
        try:
            f = open(observer_script, "r")
            observer_class = (
                f.readlines()[-1].split("--objective-name=")[1].split(" --port")[0]
            )
            f.close()
            observer = dynamically_instantiate(observer_class)
        except Exception:
            if not quiet:
                print("poli 🧪: attempting isolated observer instantiation.")
            observer = ExternalObserver(observer_name=observer_name)
    return observer



"""
This module implements Ehrlich functions as black boxes in poli.

Ehrlich functions were proposed by Stanton et al. [1] as a quick-and-easy
alternative for testing discrete sequence optimizers (with protein
optimization in mind). They are deviced to

(i) be easy to query,
(ii) have feasible and unfeasible sequences,
(iii) have uninformative random samples (i.e. randomly sampling
     and evaluating should not be competitive, as many of these
     should be unfeasible).
(iv) be maximized when certain motifs are present in the sequence.
     These motifs can be long-range within the sequence, and are
     meant to be non-additive.

Check the references for details on the implementation.

References
----------
[1] Stanton, S., Alberstein, R., Frey, N., Watkins, A., & Cho, K. (2024).
    Closed-Form Test Functions for Biophysical Sequence Optimization Algorithms.
    arXiv preprint arXiv:2407.00236. https://arxiv.org/abs/2407.00236
"""

from __future__ import annotations

import warnings

import numpy as np

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.problem import Problem
from poli.core.util.proteins.defaults import AMINO_ACIDS
from poli.core.util.seeding import seed_python_numpy_and_torch
from poli.objective_repository.ehrlich._construct_feasibility_matrix import (
    _construct_transition_matrix,
)


class EhrlichBlackBox(AbstractBlackBox):
    """
    Ehrlich functions were proposed by Stanton et al. [1] as a quick-and-easy
    alternative for testing discrete sequence optimizers (with protein
    optimization in mind). They are deviced to

    (i) be easy to query,
    (ii) have feasible and unfeasible sequences,
    (iii) have uninformative random samples (i.e. randomly sampling and evaluating should not be competitive, as many of these should be unfeasible).
    (iv) be maximized when certain motifs are present in the sequence. These motifs can be long-range within the sequence, and are meant to be non-additive.

    Check the references for details on the implementation.

    Parameters
    ----------
    sequence_length : int
        The length of the sequence to be optimized. This length is fixed, and
        _only_ sequences of this length are considered.
    motif_length : int
        The length of the motifs.
    n_motifs : int
        The number of motifs.
    quantization : int, optional
        The quantization parameter. This parameter must be between 1 and the
        motif length, and the motif length must be divisible by the quantization.
        By default, it is None (which corresponds to the motif length).
    seed : int, optional
        The seed for the random number generator. By default, it is None
        (i.e. no seed is set).
    return_value_on_unfeasible : float, optional
        The value to be returned when an unfeasible sequence is evaluated.
        By default, it is -np.inf.
    feasibility_matrix_temperature : float, optional
        The temperature parameter for the feasibility matrix's softmax. By
        default, it is 0.5.
    feasibility_matrix_band_length : int, optional
        The band length for the non-zero values in the feasibility matrix.
        By default, it is None (i.e. if the alphabet size is v, the band
        length is v - 2 * (v // 5)).
    alphabet : list of str, optional
        The alphabet to be used for the sequences. By default, it is the
        of 20 amino acids.
    batch_size : int, optional
        The batch size for the black box. By default, it is None (i.e. all
        sequences are evaluated in a vectorized way).
    parallelize : bool, optional
        Whether to parallelize the evaluation of the black box. By default,
        it is False.
    num_workers : int, optional
        The number of processors used in parallelization.
    evaluation_budget : int, optional
        The evaluation budget for the black box. By default, it is infinite.

    References
    ----------
    [1] Stanton, S., Alberstein, R., Frey, N., Watkins, A., & Cho, K. (2024).
        Closed-Form Test Functions for Biophysical Sequence Optimization Algorithms.
        arXiv preprint arXiv:2407.00236. https://arxiv.org/abs/2407.00236

    """

    def __init__(
        self,
        sequence_length: int,
        motif_length: int,
        n_motifs: int,
        quantization: int | None = None,
        seed: int = None,
        return_value_on_unfeasible: float = -np.inf,
        feasibility_matrix_temperature: float = 0.5,
        feasibility_matrix_band_length: int | None = None,
        alphabet: list[str] = AMINO_ACIDS,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
    ):
        warnings.warn(
            "This EhrlichBlackBox class is different from the original "
            "implementation provided by Stanton et al. If you are interested in "
            "their implementation (for exact comaprisons), please use "
            "EhrlichHoloBlackBox after installing with pip install poli-core[ehrlich].",
            UserWarning,
        )

        super().__init__(batch_size, parallelize, num_workers, evaluation_budget)
        self.alphabet = alphabet
        self.sequence_length = sequence_length
        self.return_value_on_unfeasible = return_value_on_unfeasible

        if seed is not None:
            seed_python_numpy_and_torch(seed)

        if motif_length * n_motifs > sequence_length:
            raise ValueError(
                "The total length of the motifs is greater than the sequence length."
            )

        if quantization is None:
            quantization = motif_length

        if not (1 <= quantization <= motif_length) or motif_length % quantization != 0:
            raise ValueError(
                "The quantization parameter must be between 1 and the motif length, "
                "and the motif length must be divisible by the quantization."
            )

        self.motif_length = motif_length
        self.n_motifs = n_motifs
        self.quantization = quantization

        self.transition_matrix = _construct_transition_matrix(
            size=len(alphabet),
            seed=seed,
            temperature=feasibility_matrix_temperature,
            band_length=feasibility_matrix_band_length,
        )

        self.motifs = self.construct_random_motifs(
            motif_length=motif_length,
            n_motifs=n_motifs,
            seed=seed,
        )
        self.offsets = self.construct_random_offsets(
            motif_length=motif_length,
            n_motifs=n_motifs,
            seed=seed,
        )

    def _sample_random_sequence(
        self,
        length: int | None = None,
        random_state: int | np.random.RandomState | None = None,
        repeating_allowed: bool = True,
    ) -> str:
        """
        Uses the sparse transition matrix to generate a random sequence
        of a given length.
        """
        if length is None:
            length = self.sequence_length

        if random_state is None:
            random_state = np.random.RandomState()

        if isinstance(random_state, int):
            random_state = np.random.RandomState(random_state)
        elif isinstance(random_state, np.random.RandomState):
            pass
        else:
            raise ValueError(
                "The random_state parameter must be an integer or an instance of "
                "np.random.RandomState."
            )

        sequence = self.alphabet[random_state.randint(len(self.alphabet))]
        current_state = self.alphabet.index(sequence)

        for _ in range(length - 1):
            next_state = random_state.choice(
                len(self.alphabet), p=self.transition_matrix[current_state]
            )
            if not repeating_allowed:
                while next_state == current_state:
                    next_state = random_state.choice(
                        len(self.alphabet),
                        p=self.transition_matrix[current_state],
                    )
            sequence += self.alphabet[next_state]
            current_state = next_state

        return sequence

    def _is_feasible(self, sequence: str | np.ndarray) -> bool:
        """
        Checks whether a sequence (str or array of one sequence) is feasible
        under the transition matrix. This is done by looping through
        the sequence and determining whether the transition probabilities
        are non-zero.
        """
        if isinstance(sequence, np.ndarray):
            assert sequence.ndim == 1 or sequence.shape[0] == 1
            sequence = "".join(sequence.flatten())

        current_state = self.alphabet.index(sequence[0])
        for i in range(1, len(sequence)):
            next_state = self.alphabet.index(sequence[i])

            if self.transition_matrix[current_state, next_state] == 0.0:
                return False
            current_state = next_state

        return True

    def construct_random_motifs(
        self, motif_length: int, n_motifs: int, seed: int = None
    ) -> np.ndarray:
        """
        Creates a given number of random motifs of a certain length.
        """
        assert motif_length * n_motifs <= self.sequence_length

        random_state = np.random.RandomState(seed)

        # Sampling a sequence of length motif_length * n_motifs
        sequence = self._sample_random_sequence(
            length=motif_length * n_motifs,
            random_state=random_state,
            repeating_allowed=False,
        )

        # Chunking it into n_motifs
        motifs = np.array(
            [
                list(sequence[i * motif_length : (i + 1) * motif_length])
                for i in range(0, n_motifs)
            ]
        )

        return motifs

    def construct_random_offsets(
        self,
        motif_length: int,
        n_motifs: int,
        seed: int = None,
    ) -> np.ndarray:
        """
        Creates a given number of random offsets for the motifs.
        """
        all_motifs_length = motif_length * n_motifs

        # For each motif, we sample weights in the simplex
        # from a uniform dirichlet
        random_state = np.random.RandomState(seed)

        offsets = []
        for _ in range(n_motifs):
            weights = random_state.dirichlet(np.ones(motif_length - 1))
            _offset_for_motif = [0]
            for weight in weights:
                _offset_for_motif.append(
                    1
                    + np.floor(
                        weight * (self.sequence_length - all_motifs_length) // n_motifs
                    )
                )

            offset_for_motif = np.cumsum(np.array(_offset_for_motif, dtype=int))

            offsets.append(offset_for_motif)

        return np.array(offsets)

    def construct_optimal_solution(
        self, motifs: np.ndarray | None = None, offsets: np.ndarray | None = None
    ) -> np.ndarray:
        """
        Constructs an optimal solution for a given set of motifs and offsets.

        If None are provided, then the motifs and offsets of the black box
        are used.
        """
        if motifs is None:
            motifs = self.motifs

        if offsets is None:
            offsets = self.offsets

        # de-cumsum the offsets
        offsets = np.diff(offsets, prepend=0)
        optimal_sequence = []

        for motif, offset in zip(motifs, offsets):
            # We write first l - 1 characters according to the offsets,
            # and wait to write the last character at the end
            for character, next_offset_value in zip(motif, offset[1:]):
                # Put the current character in the current position all the way through just before the offset
                optimal_sequence += [character] * next_offset_value

            # Write the last character
            optimal_sequence += [motif[-1]]

        # We pad until the sequence length with the last character
        # of the last motif
        optimal_sequence += [motifs[-1][-1]] * (
            self.sequence_length - len(optimal_sequence)
        )

        return np.array(optimal_sequence).reshape(1, -1)

    def _maximal_motif_matches(
        self, sequence: np.ndarray, motif: np.ndarray, offset: np.ndarray
    ) -> int:
        """
        Counts the maximal motif match.
        """
        assert sequence.ndim == 1 or sequence.shape[0] == 1
        sequence = "".join(sequence.flatten())
        maximal_match = 0
        for seq_idx in range(len(sequence) - max(offset)):
            matches = 0
            sequence_at_offset = np.array(
                [sequence[seq_idx + offset_value] for offset_value in offset]
            )
            matches = sum(sequence_at_offset == motif)

            maximal_match = max(maximal_match, matches)

        return maximal_match

    def _black_box(self, x: np.ndarray, context=None) -> np.ndarray:
        """
        Evaluates the sequences in x by checking maximal matches and multiplying.
        """
        values = []
        for sequence in x:
            if not self._is_feasible(sequence):
                values.append(self.return_value_on_unfeasible)
                continue

            value = 1.0
            for motif, offset in zip(self.motifs, self.offsets):
                maximal_matches = self._maximal_motif_matches(sequence, motif, offset)
                value *= (
                    maximal_matches // (self.motif_length / self.quantization)
                ) / self.quantization

            values.append(value)

        return np.array(values).reshape(-1, 1)

    def get_black_box_info(self) -> BlackBoxInformation:
        return BlackBoxInformation(
            name="ehrlich",
            max_sequence_length=self.sequence_length,
            aligned=True,
            fixed_length=True,
            deterministic=True,
            alphabet=self.alphabet,
            log_transform_recommended=False,
            discrete=True,
            padding_token="",
        )


class EhrlichProblemFactory(AbstractProblemFactory):
    """
    A factory for creating Ehrlich functions and initial conditions.

    References
    ----------
    [1] Stanton, S., Alberstein, R., Frey, N., Watkins, A., & Cho, K. (2024).
        Closed-Form Test Functions for Biophysical Sequence Optimization Algorithms.
        arXiv preprint arXiv:2407.00236. https://arxiv.org/abs/2407.00236
    """

    def create(
        self,
        sequence_length: int,
        motif_length: int,
        n_motifs: int,
        quantization: int | None = None,
        seed: int = None,
        return_value_on_unfeasible: float = -np.inf,
        alphabet: list[str] = AMINO_ACIDS,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        """
        Creates an Ehrlich function problem (containing an Ehrlich black box and
        an initial condition).

        Parameters
        ----------
        sequence_length : int
            The length of the sequence to be optimized. This length is fixed, and
            _only_ sequences of this length are considered.
        motif_length : int
            The length of the motifs.
        n_motifs : int
            The number of motifs.
        quantization : int, optional
            The quantization parameter. This parameter must be between 1 and the
            motif length, and the motif length must be divisible by the quantization.
            By default, it is None (which corresponds to the motif length).
        seed : int, optional
            The seed for the random number generator. By default, it is None
            (i.e. no seed is set).
        return_value_on_unfeasible : float, optional
            The value to be returned when an unfeasible sequence is evaluated.
            By default, it is -np.inf.
        alphabet : list of str, optional
            The alphabet to be used for the sequences. By default, it is the
            of 20 amino acids.
        batch_size : int, optional
            The batch size for the black box. By default, it is None (i.e. all
            sequences are evaluated in a vectorized way).
        parallelize : bool, optional
            Whether to parallelize the evaluation of the black box. By default,
            it is False.
        num_workers : int, optional
            The number of processors used in parallelization.
        evaluation_budget : int, optional
            The evaluation budget for the black box. By default, it is infinite.

        References
        ----------
        [1] Stanton, S., Alberstein, R., Frey, N., Watkins, A., & Cho, K. (2024).
            Closed-Form Test Functions for Biophysical Sequence Optimization Algorithms.
            arXiv preprint arXiv:2407.00236. https://arxiv.org/abs/2407.00236
        """
        warnings.warn(
            "The EhrlichProblemFactory class is deprecated and will be removed in a future version. "
            " Please use EhrlichHoloProblemFactory after installing with pip install poli-core[ehrlich].",
            DeprecationWarning,
        )
        if seed is not None:
            seed_python_numpy_and_torch(seed)

        f = EhrlichBlackBox(
            sequence_length=sequence_length,
            motif_length=motif_length,
            n_motifs=n_motifs,
            quantization=quantization,
            seed=seed,
            return_value_on_unfeasible=return_value_on_unfeasible,
            alphabet=alphabet,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )
        x0 = np.array([list(f._sample_random_sequence())])

        return Problem(f, x0)



"""
Implements a multi_objective version of black box functions,
by this we mean simply concatenating the results of evaluating individual
objective functions.
"""

from typing import List

import numpy as np

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.black_box_information import BlackBoxInformation


class MultiObjectiveBlackBox(AbstractBlackBox):
    """
    This class implements a multi-objective black box function
    by concatenating the results of evaluating individual
    objective functions.

    Parameters
    -----------
    batch_size : int, optional
        The batch size for evaluating the black box function. Defaults to None.
    objective_functions : List[AbstractBlackBox], required
        The list of objective functions to be evaluated. Defaults to None.

    Attributes
    ----------
    objective_functions : List[AbstractBlackBox]
        The list of objective functions to be evaluated.

    Methods
    -------
    _black_box(x, context=None)
        Evaluate the black box function for the given input.

    Raises
    ------
    ValueError
        If objective_functions is not provided as a list of AbstractBlackBox
        instances or inherited classes.
    """

    # TODO: Should we evaluate in parallel?
    # TODO: At some point, we should consider unions of ProblemSetupInformations.
    #       That way, we could simply specify the list of black box functions, and
    #       then "join" their problem setups.

    def __init__(
        self,
        objective_functions: List[AbstractBlackBox],
        batch_size: int = None,
    ) -> None:
        """
        Initialize the MultiObjectiveBlackBox class.

        Parameters
        -----------
        objective_functions : List[AbstractBlackBox]
            The list of objective functions.
        batch_size : int, optional
            The batch size. Defaults to None.

        Raises
        -------
        ValueError:
            If objective_functions is not provided as a list of AbstractBlackBox instances or inherited classes.
        """
        if objective_functions is None:
            raise ValueError(
                "objective_functions must be provided as a list of AbstractBlackBox instances or inherited classes."
            )

        super().__init__(
            batch_size=batch_size,
        )

        self.objective_functions = objective_functions

    def _black_box(self, x: np.ndarray, context=None) -> np.ndarray:
        """
        Evaluate the black box function for a given input.

        Parameters
        -----------
        x : np.ndarray
            The input values to evaluate the black box function.
        context : any, optional
            Additional context information for the evaluation.

        Returns
        --------
            array-like: The concatenated results of evaluating the objective functions on the input.
        """
        # TODO: Decide on whether we want to evaluate on parallel or not.
        res = []
        for obj_function in self.objective_functions:
            res.append(obj_function(x, context))

        return np.concatenate(res, axis=1)

    def __str__(self) -> str:
        return f"MultiObjectiveBlackBox(black_boxes={self.objective_functions})"

    def __repr__(self) -> str:
        return f"<MultiObjectiveBlackBox(black_boxes={self.objective_functions}, batch_size={self.batch_size})>"

    @property
    def info(self) -> BlackBoxInformation:
        """
        Return the problem setup information for the multi-objective black box.

        Returns
        -------
        BlackBoxInformation:
            Information about the first objective function.
        """
        # TODO: what should this return, actually?
        # I'd say that we expect all objective functions to be able
        # to take the same input.
        return self.objective_functions[0].info



"""
This module implements LogP in _exactly_ the same was
as LaMBO does it [1]. We do this by importing the function
they use to compute `logp`.

References
----------
[1] “Accelerating Bayesian Optimization for Biological Sequence Design with Denoising Autoencoders.”
Stanton, Samuel, Wesley Maddox, Nate Gruver, Phillip Maffettone,
Emily Delaney, Peyton Greenside, and Andrew Gordon Wilson.
arXiv, July 12, 2022. http://arxiv.org/abs/2203.12742.
"""

from typing import Literal, Tuple

import numpy as np

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.problem import Problem
from poli.core.util.isolation.instancing import get_inner_function
from poli.core.util.seeding import seed_python_numpy_and_torch


class PenalizedLogPLamboBlackBox(AbstractBlackBox):
    """
    A black box objective function that returns the
    penalized logP of a molecule, using the same function
    that LaMBO [1] uses.

    In particular, they adjust the penalized logP using
    some "magic numbers", which are the empirical means
    and standard deviations of the dataset.
    """

    def __init__(
        self,
        string_representation: Literal["SMILES", "SELFIES"] = "SMILES",
        penalized: bool = True,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ):
        super().__init__(
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )
        from_smiles = string_representation.upper() == "SMILES"
        self.from_smiles = from_smiles
        self.penalized = penalized
        self.force_isolation = force_isolation

        # Testing whether we can import in isolation
        # to eagerly throw any import errors
        _ = get_inner_function(
            isolated_function_name="penalized_logp_lambo__isolated",
            class_name="PenalizedLogPIsolatedLogic",
            module_to_import="poli.objective_repository.penalized_logp_lambo.isolated_function",
            force_isolation=force_isolation,
            from_smiles=from_smiles,
            penalized=penalized,
        )

    def _black_box(self, x: np.ndarray, context: dict = None):
        """
        Assuming that x is an array of strings (of shape [b,L]),
        we concatenate, translate to smiles if it's
        necessary, and then computes the penalized logP.

        If the inputs are SELFIES, it translates first to SMILES,
        and then computes the penalized logP. If the translation
        threw an error, we return NaN instead.
        """
        inner_function = get_inner_function(
            isolated_function_name="penalized_logp_lambo__isolated",
            class_name="PenalizedLogPIsolatedLogic",
            module_to_import="poli.objective_repository.penalized_logp_lambo.isolated_function",
            force_isolation=self.force_isolation,
            from_smiles=self.from_smiles,
            penalized=self.penalized,
        )
        return inner_function(x, context)

    def get_black_box_info(self) -> BlackBoxInformation:
        return BlackBoxInformation(
            name="penalized_logp_lambo",
            max_sequence_length=np.inf,
            aligned=False,
            fixed_length=False,
            alphabet=None,  # TODO: add when we settle for an alphabet
            deterministic=True,
            log_transform_recommended=False,
            discrete=True,
            fidelity=None,
            padding_token="",
        )


class PenalizedLogPLamboProblemFactory(AbstractProblemFactory):
    def create(
        self,
        penalized: bool = True,
        string_representation: str = "SMILES",
        seed: int = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Tuple[AbstractBlackBox, np.ndarray, np.ndarray]:
        if seed is not None:
            seed_python_numpy_and_torch(seed)

        if string_representation.upper() not in ["SMILES", "SELFIES"]:
            raise ValueError(
                "Missing required keyword argument: string_representation: str. "
                "String representation must be either 'SMILES' or 'SELFIES'."
            )

        f = PenalizedLogPLamboBlackBox(
            string_representation=string_representation.upper(),
            penalized=penalized,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
            force_isolation=force_isolation,
        )

        if string_representation.upper() == "SMILES":
            x0 = np.array([["C"]])
        else:
            x0 = np.array([["[C]"]])

        problem = Problem(f, x0)

        return problem



"""Utilities for seeding random number generators."""

from .seeding import seed_numpy, seed_python, seed_python_numpy_and_torch, seed_torch

__all__ = ["seed_numpy", "seed_python", "seed_python_numpy_and_torch", "seed_torch"]



from pathlib import Path

import numpy as np
import pytest

from poli import objective_factory
from poli.objective_repository import (
    AlbuterolSimilarityBlackBox,
    AmlodipineMPOBlackBox,
    CelecoxibRediscoveryBlackBox,
    DecoHopBlackBox,
    DRD2BlackBox,
    FexofenadineMPOBlackBox,
    GSK3BetaBlackBox,
    IsomerC7H8N2O2BlackBox,
    IsomerC9H10N2O2PF2ClBlackBox,
    JNK3BlackBox,
    Median1BlackBox,
    Median2BlackBox,
    MestranolSimilarityBlackBox,
    OsimetrinibMPOBlackBox,
    PerindoprilMPOBlackBox,
    RanolazineMPOBlackBox,
    SABlackBox,
    ScaffoldHopBlackBox,
    SitagliptinMPOBlackBox,
    ThiothixeneRediscoveryBlackBox,
    TroglitazoneRediscoveryBlackBox,
    ValsartanSMARTSBlackBox,
    ZaleplonMPOBlackBox,
)

THIS_DIR = Path(__file__).parent.resolve()

SEED = np.random.randint(0, 1000)


@pytest.mark.poli__lambo
def test_penalized_logp_lambo():
    """
    Testing whether we can register the logp problem
    from lambo.
    """

    _ = pytest.importorskip("lambo")

    # Using create
    _ = objective_factory.create(name="penalized_logp_lambo")


@pytest.mark.poli__dockstring
def test_querying_dockstring_using_smiles():
    """
    In this test, we force-register and query dockstring.
    """

    problem = objective_factory.create(
        name="dockstring",
        target_name="DRD2",
        string_representation="SMILES",
    )
    f = problem.black_box

    # Docking another smiles
    x1 = np.array([list("CC(=O)OC1=CC=CC=C1C(=O)O")])
    _ = f(x1)

    f.terminate()


@pytest.mark.poli__dockstring
def test_querying_dockstring_using_selfies():
    """
    In this test, we check whether dockstring still
    works when using SELFIES instead of SMILES.
    """

    problem = objective_factory.create(
        name="dockstring",
        target_name="ABL1",
        string_representation="SELFIES",
    )
    f = problem.black_box

    # Docking another smiles
    selfies_aspirin = np.array(
        [
            [
                "[C]",
                "[C]",
                "[=Branch1]",
                "[C]",
                "[=O]",
                "[O]",
                "[C]",
                "[=C]",
                "[C]",
                "[=C]",
                "[C]",
                "[=C]",
                "[Ring1]",
                "[=Branch1]",
                "[C]",
                "[=Branch1]",
                "[C]",
                "[=O]",
                "[O]",
            ]
        ]
    )

    _ = f(selfies_aspirin)
    f.terminate()


test_data_for_pmo = [
    ("gsk3_beta", GSK3BetaBlackBox, {"string_representation": "SMILES"}, 0.03),
    (
        "drd2_docking",
        DRD2BlackBox,
        {"string_representation": "SMILES"},
        0.0015465365340340924,
    ),
    (
        "sa_tdc",
        SABlackBox,
        {"string_representation": "SMILES"},
        2.706977149048555,
    ),
    (
        "celecoxib_rediscovery",
        CelecoxibRediscoveryBlackBox,
        {"string_representation": "SMILES"},
        0.14728682170542637,
    ),
    (
        "thiothixene_rediscovery",
        ThiothixeneRediscoveryBlackBox,
        {"string_representation": "SMILES"},
        0.17391304347826086,
    ),
    (
        "troglitazone_rediscovery",
        TroglitazoneRediscoveryBlackBox,
        {"string_representation": "SMILES"},
        0.24427480916030533,
    ),
    (
        "albuterol_similarity",
        AlbuterolSimilarityBlackBox,
        {"string_representation": "SMILES"},
        0.2772277227722772,
    ),
    (
        "mestranol_similarity",
        MestranolSimilarityBlackBox,
        {"string_representation": "SMILES"},
        0.19460880999342536,
    ),
    (
        "amlodipine_mpo",
        AmlodipineMPOBlackBox,
        {"string_representation": "SMILES"},
        0.461083967620704,
    ),
    (
        "fexofenadine_mpo",
        FexofenadineMPOBlackBox,
        {"string_representation": "SMILES"},
        0.4336446174984538,
    ),
    (
        "osimetrinib_mpo",
        OsimetrinibMPOBlackBox,
        {"string_representation": "SMILES"},
        0.09011742702110873,
    ),
    (
        "perindopril_mpo",
        PerindoprilMPOBlackBox,
        {"string_representation": "SMILES"},
        0.36023741111440966,
    ),
    (
        "ranolazine_mpo",
        RanolazineMPOBlackBox,
        {"string_representation": "SMILES"},
        0.29285467466584664,
    ),
    # The following two have a discrepancy with
    # the TDC docs. An issue has been raised (#244)
    (
        "sitagliptin_mpo",
        SitagliptinMPOBlackBox,
        {"string_representation": "SMILES"},
        3.34970667598234e-12,
    ),
    (
        "zaleplon_mpo",
        ZaleplonMPOBlackBox,
        {"string_representation": "SMILES"},
        0.0019017991803329235,
    ),
    # The following should actually output 0.5338365434669443,
    # but TDC has a discrepancy. An issue has been raised (#244)
    (
        "deco_hop",
        DecoHopBlackBox,
        {"string_representation": "SMILES"},
        0.5338365434669443,
    ),
    (
        "scaffold_hop",
        ScaffoldHopBlackBox,
        {"string_representation": "SMILES"},
        0.38446411012782694,
    ),
    # The following two have a discrepancy with
    # the TDC docs. An issue has been raised (#244)
    (
        "isomer_c7h8n2o2",
        IsomerC7H8N2O2BlackBox,
        {"string_representation": "SMILES"},
        2.1987591132394053e-34,
    ),
    (
        "isomer_c9h10n2o2pf2cl",
        IsomerC9H10N2O2PF2ClBlackBox,
        {"string_representation": "SMILES"},
        1.713908431542013e-15,
    ),
    (
        "median_1",
        Median1BlackBox,
        {"string_representation": "SMILES"},
        0.09722243533981723,
    ),
    (
        "median_2",
        Median2BlackBox,
        {"string_representation": "SMILES"},
        0.12259690287307903,
    ),
    (
        "valsartan_smarts",
        ValsartanSMARTSBlackBox,
        {"string_representation": "SMILES"},
        0.0,
    ),
    ("jnk3", JNK3BlackBox, {"string_representation": "SMILES"}, 0.01),
]


@pytest.mark.poli__tdc
@pytest.mark.parametrize(
    "black_box_name, black_box_class, kwargs_for_black_box, value_to_check",
    test_data_for_pmo,
)
def test_pmo_black_boxes(
    black_box_name, black_box_class, kwargs_for_black_box, value_to_check
):
    from poli import create
    from poli.core.util.seeding import seed_python_numpy_and_torch

    problem = create(
        name=black_box_name,
        seed=SEED,
        **kwargs_for_black_box,
    )
    x0 = problem.x0
    y0 = problem.black_box(x0)

    seed_python_numpy_and_torch(seed=SEED)
    f = black_box_class(**kwargs_for_black_box)
    y0_ = f(x0)

    assert np.allclose(y0_, y0)
    if value_to_check is not None:
        assert (y0_ == value_to_check).all()



"""Dockstring objective factory and function.

This module implements and registers a black box objective function
(and a repository) for dockstring [1], which is a simple API for
assessing the docking score of a small molecule to a given protein.

References
----------
[1] “DOCKSTRING: Easy Molecular Docking Yields Better Benchmarks for Ligand Design.”
    García-Ortegón, Miguel, Gregor N. C. Simm, Austin J. Tripp,
    José Miguel Hernández-Lobato, Andreas Bender, and Sergio Bacallado.
    Journal of Chemical Information and Modeling 62, no. 15 (August 8, 2022): 3486-3502.
    https://doi.org/10.1021/acs.jcim.1c01334.
"""

from __future__ import annotations

from typing import Literal

import numpy as np
import selfies as sf

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.problem import Problem
from poli.core.util.chemistry.string_to_molecule import translate_smiles_to_selfies
from poli.core.util.isolation.instancing import get_inner_function
from poli.core.util.seeding import seed_python_numpy_and_torch


class DockstringBlackBox(AbstractBlackBox):
    """
    Black box implementation for the Dockstring problem.

    Dockstring is a simple API for assessing the docking score
    of a small molecule to a given protein [1].

    Parameters
    ----------
    target_name : str
        The name of the target protein.
    string_representation : str, optional
        The string representation of the molecules. Either SMILES or SELFIES.
        Default is SMILES.
    batch_size : int, optional
        The batch size for processing multiple inputs simultaneously, by default None.
    parallelize : bool, optional
        Flag indicating whether to parallelize the computation, by default False.
    num_workers : int, optional
        The number of workers to use for parallel computation, by default None.
    evaluation_budget:  int, optional
        The maximum number of function evaluations. Default is infinity.

    Attributes
    ----------
    alphabet : dict
        The mapping of symbols to their corresponding indices in the alphabet.

    Methods
    -------
    _black_box(x, context=None)
        The black box function.

    References
    ----------
    [1] “DOCKSTRING: Easy Molecular Docking Yields Better Benchmarks for Ligand Design.”
        García-Ortegón, Miguel, Gregor N. C. Simm, Austin J. Tripp,
        José Miguel Hernández-Lobato, Andreas Bender, and Sergio Bacallado.
        Journal of Chemical Information and Modeling 62, no. 15 (August 8, 2022): 3486-3502.
        https://doi.org/10.1021/acs.jcim.1c01334.
    """

    def __init__(
        self,
        target_name: str,
        string_representation: Literal["SMILES", "SELFIES"] = "SMILES",
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ):
        """
        Initialize the dockstring black box object.

        Parameters
        ----------
        target_name : str
            The name of the target protein.
        string_representation : str
            The string representation of the molecules. Either SMILES or SELFIES.
            Default is SMILES.
        batch_size : int, optional
            The batch size for processing data, by default None.
        parallelize : bool, optional
            Flag indicating whether to parallelize the processing, by default False.
        num_workers : int, optional
            The number of workers to use for parallel processing, by default None.
        evaluation_budget:  int, optional
            The maximum number of function evaluations. Default is infinity.
        """
        assert (
            target_name is not None
        ), "Missing required keyword argument 'target_name'. "

        super().__init__(
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )
        self.target_name = target_name
        self.string_representation = string_representation
        self.force_isolation = force_isolation

        # Testing whether we can create the isolated function
        # without raising an exception.
        _ = get_inner_function(
            isolated_function_name="dockstring__isolated",
            class_name="IsolatedDockstringFunction",
            module_to_import="poli.objective_repository.dockstring.isolated_function",
            force_isolation=force_isolation,
            target_name=target_name,
            string_representation=string_representation,
        )

    def _black_box(self, x: np.ndarray, context=None) -> np.ndarray:
        """Evaluating the black box function.

        Parameters
        ----------
        x : np.ndarray
            A molecule represented as a string of shape [b, L],
            where b is the batch size and L is the length of the string.
            We expect the elements in a row of x to be the tokens
            of a molecule string representation.
        context : any, optional
            Additional context information for the evaluation. Defaults to None.

        Returns
        -------
        y: np.ndarray
            The output of the black box function.

        Raises
        ------
        AssertionError
            If the input is not a 2D array of strings.
        Exception
            If the docking score cannot be computed.
        """
        inner_function = get_inner_function(
            isolated_function_name="dockstring__isolated",
            class_name="IsolatedDockstringFunction",
            module_to_import="poli.objective_repository.dockstring.isolated_function",
            force_isolation=self.force_isolation,
            quiet=True,
            target_name=self.target_name,
            string_representation=self.string_representation,
        )
        return inner_function(x, context=context)

    def get_black_box_info(self) -> BlackBoxInformation:
        return BlackBoxInformation(
            name="dockstring",
            max_sequence_length=np.inf,
            aligned=False,
            fixed_length=False,
            deterministic=True,
            alphabet=None,  # TODO: fix when we have a smiles alphabet
            log_transform_recommended=False,
            discrete=True,
            padding_token="",
        )


class DockstringProblemFactory(AbstractProblemFactory):
    """Problem factory for the Dockstring problem.

    Dockstring is a simple API for assessing the docking score
    of a small molecule to a given protein [1].

    Methods
    -------
    create(...)
        Creates a problem instance with the specified parameters.

    References
    ----------
    [1] “DOCKSTRING: Easy Molecular Docking Yields Better Benchmarks for Ligand Design.”
        García-Ortegón, Miguel, Gregor N. C. Simm, Austin J. Tripp,
        José Miguel Hernández-Lobato, Andreas Bender, and Sergio Bacallado.
        Journal of Chemical Information and Modeling 62, no. 15 (August 8, 2022): 3486-3502.
        https://doi.org/10.1021/acs.jcim.1c01334.
    """

    def create(
        self,
        target_name: str,
        string_representation: Literal["SMILES", "SELFIES"] = "SMILES",
        seed: int = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        """Creates a dockstring black box function and initial observations.

        Parameters
        ----------
        target_name : str
            The name of the target protein (see dockstring for more details).
        string_representation : str, optional
            The string representation of the molecules. Either SMILES or SELFIES.
            Default is SMILES.
        seed : int, optional
            The seed value for random number generation, by default None.
        batch_size : int, optional
            The batch size for processing data, by default None.
        parallelize : bool, optional
            Flag indicating whether to parallelize the processing, by default False.
        num_workers : int, optional
            The number of workers to use for parallel processing, by default we
            use half the number of available CPUs (rounded down).
        evaluation_budget:  int, optional
            The maximum number of function evaluations. Default is infinity.

        Returns
        -------
        problem: Problem
            A problem definition with the dockstring black box function and initial observations.
        """
        assert (
            target_name is not None
        ), "Missing required keyword argument 'target_name'. "

        if seed is not None:
            seed_python_numpy_and_torch(seed)

        dockstring_black_box = DockstringBlackBox(
            target_name=target_name,
            string_representation=string_representation,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
            force_isolation=force_isolation,
        )

        # Using the initial example they provide in the
        # dockstring documentation (i.e. DRD2 and Risperidone)
        risperidone_smiles = "CC1=C(C(=O)N2CCCCC2=N1)CCN3CCC(CC3)C4=NOC5=C4C=CC(=C5)F"
        if string_representation.upper() == "SMILES":
            # TODO: replace for proper smiles tokenization.
            x0 = np.array([list(risperidone_smiles)])
        elif string_representation.upper() == "SELFIES":
            risperidone_selfies = translate_smiles_to_selfies([risperidone_smiles])[0]
            risperidone_selfies_as_tokens = list(sf.split_selfies(risperidone_selfies))
            x0 = np.array([risperidone_selfies_as_tokens])
        else:
            raise ValueError(
                f"Invalid string representation. Expected SMILES or SELFIES but received {string_representation}."
            )

        dockstring_problem = Problem(black_box=dockstring_black_box, x0=x0)

        return dockstring_problem



import argparse
import os
import sys
import time

import Bio
import Bio.PDB
import numpy as np
import openmm.app
import simtk.unit

from . import grid


def extract_atomic_features(pdb_filename):
    """Extract atomic features from pdb"""

    # Parse structure with biopython
    pdb_parser = Bio.PDB.PDBParser()
    pdb_id = os.path.basename(pdb_filename).split(".")[0]
    structure = pdb_parser.get_structure(pdb_id, pdb_filename)
    first_model = structure.get_list()[0]
    ppb = Bio.PDB.PPBuilder()

    sequence = []  # Only used for an assertion / sanity check
    sequence_onehot = []
    resids_pdb = []
    chain_ids = []
    for i, chain in enumerate(first_model):
        # Append chainid
        chain_ids.append(chain.id)

        # Sequence of residue names for this chain
        sequence_chain = []
        for res in chain.get_residues():
            resids_pdb.append(res.id[1])
            sequence_chain.append(res.resname.strip())

        # Add to global container for this protein
        sequence.append(sequence_chain)

        # Convert residue names to amino acid indices
        aa_indices = []
        for aa in sequence_chain:
            try:
                aa_index = Bio.PDB.Polypeptide.three_to_index(aa)
            except:
                aa_index = 20
            aa_indices.append(aa_index)

        # Convert to one-hot encoding
        aa_onehot_chain = np.zeros((len(aa_indices), 21))
        aa_onehot_chain[np.arange(len(aa_indices)), aa_indices] = 1
        sequence_onehot.append(aa_onehot_chain)

    # Keep track of boundaries of individual chains
    chain_boundary_indices = np.cumsum([0] + [len(entry) for entry in sequence_onehot])

    # Collapse all chain segments into one. The individual chains
    # will be recoverable through chain_boundary_indices
    sequence_onehot = np.concatenate(sequence_onehot)

    # Convert resids of pdb to arr
    resids_pdb = np.array(resids_pdb)

    # Extract positions using OpenMM.
    pdb_simtk = openmm.app.PDBFile(pdb_filename)
    positions = pdb_simtk.getPositions()

    # Save features in a dictionary
    features = {}
    features["atom_names"] = []  # Atom names
    features["res_indices"] = (
        []
    )  # Global res indices (no reset across chains and starts from 0)
    features["resids_pdb"] = []  # Resids in the actual odbs
    features["x"] = []
    features["y"] = []
    features["z"] = []

    # Iterate over chain,residue,atoms and extract features
    for i, chain in enumerate(pdb_simtk.getTopology().chains()):
        chain_start_index = chain_boundary_indices[i]
        for j, residue in enumerate(chain.residues()):
            for atom in residue.atoms():
                # Extract atom features
                index = atom.index
                position = list(positions[index].value_in_unit(simtk.unit.angstrom))
                features["atom_names"].append(atom.name)
                features["res_indices"].append(residue.index)
                features["x"].append(position[0])
                features["y"].append(position[1])
                features["z"].append(position[2])

                # Sanity check
                residue_index_local = residue.index - chain_start_index
                assert residue.name == sequence[i][residue_index_local]

    # Convert valid lists to numpy arrays
    # (even convert atom_names since its simpler to mask with despite being str)
    features["atom_names"] = np.array(features["atom_names"], dtype="a5")
    features["res_indices"] = np.array(features["res_indices"], dtype=np.int32)
    features["x"] = np.array(features["x"], dtype=np.float32)
    features["y"] = np.array(features["y"], dtype=np.float32)
    features["z"] = np.array(features["z"], dtype=np.float32)

    return features, sequence_onehot, chain_ids, chain_boundary_indices, resids_pdb


def extract_coordinates(features, max_radius, include_center):
    """Extract environment coordinates within specifies crieteria"""

    # Extract coordinates as normal numpy array
    position_array = np.vstack([features["x"], features["y"], features["z"]]).T

    # Retrieve residue indices as numpy int array
    # This array has many repeats, since it follows the sequence of atoms,
    # not residues. It counts globally across chains, i.e. no resets and
    # starts from zero.
    res_indices_glob = features["res_indices"]
    res_indices_uniq = np.unique(
        res_indices_glob
    )  # Has length == number of total residues

    # Begin
    selector_list = []
    indices_list = []
    xyz_ref_origo_list = []
    for residue_index in res_indices_uniq:
        # Extract origin mask
        if (
            np.logical_and(
                res_indices_glob == residue_index, features["atom_names"] == b"N"
            ).any()
            and np.logical_and(
                res_indices_glob == residue_index, features["atom_names"] == b"CA"
            ).any()
            and np.logical_and(
                res_indices_glob == residue_index, features["atom_names"] == b"C"
            ).any()
        ):
            N_mask = np.logical_and(
                res_indices_glob == residue_index, features["atom_names"] == b"N"
            )
            CA_mask = np.logical_and(
                res_indices_glob == residue_index, features["atom_names"] == b"CA"
            )
            C_mask = np.logical_and(
                res_indices_glob == residue_index, features["atom_names"] == b"C"
            )
        else:
            # Store None to maintain indices
            indices_list.append(None)
            selector_list.append(None)
            continue

        # Extract origin
        pos_N = np.array(
            [features["x"][N_mask], features["y"][N_mask], features["z"][N_mask]]
        ).squeeze()
        pos_CA = np.array(
            [features["x"][CA_mask], features["y"][CA_mask], features["z"][CA_mask]]
        ).squeeze()
        pos_C = np.array(
            [features["x"][C_mask], features["y"][C_mask], features["z"][C_mask]]
        ).squeeze()

        # Define Cartesian coordinate system
        coordinate_system = grid.CoordinateSystem["cartesian"]
        z_direction = grid.ZDirection[grid.ZDirection.outward.name]

        # Define local coordinate system
        rot_matrix = grid.define_coordinate_system(pos_N, pos_CA, pos_C, z_direction)

        # Calculate coordinates relative to origin
        xyz = position_array - pos_CA

        # Rotate to the local reference
        xyz = np.dot(rot_matrix, xyz.T).T

        # Calculate radius
        r = np.sqrt(xyz[:, 0] ** 2 + xyz[:, 1] ** 2 + xyz[:, 2] ** 2)

        if include_center:
            selector = np.where(r < max_radius)[0]
        else:
            # ALSO exclude features from residue itself
            selector = np.where(
                np.logical_and(r < max_radius, res_indices_glob != residue_index)
            )[0]

        xyz_ref_origo_list.append(xyz[selector])
        selector_list.append(selector)

    # Find max number of atoms in an environment with radias = max_radius
    max_selector = max(
        [len(selector) for selector in selector_list if selector is not None]
    )
    selector_array = np.full((len(selector_list), max_selector), -1, dtype=np.int32)
    for i, selector in enumerate(selector_list):
        if selector is not None:
            selector_array[i, : len(selector)] = selector.astype(np.int32)

    # Output actual coordinates and atom types rather than embedding in a grid.
    atom_type_list = ["C", "N", "O", "H", "S", "P"]
    atom_types_numeric = np.array(
        [atom_type_list.index(x.decode("utf-8")[0]) for x in features["atom_names"]]
    )  # Zero refers to the first letter of atom name

    # Save selected coordinates in array with shape defined by the max number of atoms in any of environment
    xyz_ref_origo_arr = np.full(
        shape=[len(selector_list), max_selector, 3],
        fill_value=[-99, -99, -99],
        dtype=np.float32,
    )
    for i, xyz_selected in enumerate(xyz_ref_origo_list):
        if xyz_selected is not None:
            xyz_ref_origo_arr[i, : xyz_selected.shape[0], :] = xyz_selected

    return xyz_ref_origo_arr, atom_types_numeric, selector_array


def extract_environments(
    pdb_filename: str,
    pdb_id: str,
    max_radius: float = 9.0,
    out_dir: str = "./",
    include_center: bool = False,
):
    """
    Extract residue environments from PDB file. Outputs .npz file.

    Parameters
    ----------
    pdb_filename: str
        PDB filename to extract environments from
    pdb_id: str
        PDBID. Used as a prefix for the output file, and does not have to follow
        the standard 4 character nomenclature
    max_radiues: float
        Max radius from center CA atom in Angstrom
    include_center: bool
        Whether to include the center residue. For the cavity model, this only
        makes sense with set to False, since we are classifying the missing
        center residue.
    """

    # Extract atomic features and other relevant info
    (
        features,
        sequence_onehot,
        chain_ids,
        chain_boundary_indices,
        resids_pdb,
    ) = extract_atomic_features(pdb_filename)

    # Extract relevant coordinates (already masked with selector and referenced),
    # all atom types and the mask for each residue (selector_array)
    xyz_ref_origo_arr, atom_types_numeric, selector_array = extract_coordinates(
        features, max_radius, include_center
    )

    # Save as .npz
    np.savez_compressed(
        out_dir + f"/{pdb_id}_coordinate_features",
        atom_types_numeric=atom_types_numeric,
        positions=xyz_ref_origo_arr,
        selector=selector_array,
        aa_onehot=sequence_onehot,
        chain_boundary_indices=chain_boundary_indices,
        chain_ids=chain_ids,
        residue_numbers=resids_pdb,
    )


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


if __name__ == "__main__":
    t0 = time.time()

    # Argument parser
    parser = argparse.ArgumentParser()
    parser.add_argument("--pdb_in", type=str)
    parser.add_argument("--max_radius", type=float, default=9.0)
    parser.add_argument("--include_center", type=str2bool, default=False)
    parser.add_argument("--out_dir", type=str, default="./")
    args_dict = vars(parser.parse_args())

    # Settings
    pdb_filename = args_dict["pdb_in"]
    pdb_id = os.path.basename(pdb_filename).split(".")[0]
    max_radius = args_dict["max_radius"]
    out_dir = args_dict["out_dir"]
    include_center = args_dict["include_center"]

    # Extract
    extract_environments(pdb_filename, pdb_id, max_radius, out_dir, include_center)
    t1 = time.time()
    sys.stdout.flush()
    print(f"Time for parsing environments from {pdb_filename}: {t1-t0}")



import numpy as np

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.problem import Problem
from poli.core.util.isolation.instancing import instance_function_as_isolated_process
from poli.core.util.proteins.defaults import AMINO_ACIDS
from poli.core.util.seeding import seed_python_numpy_and_torch


class GFPSelectionBlackBox(AbstractBlackBox):
    def __init__(
        self,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        seed: int = None,
        force_isolation: bool = False,
    ):
        super().__init__(
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )
        if not force_isolation:
            try:
                from poli.objective_repository.gfp_select.isolated_function import (
                    GFPSelectIsolatedLogic,
                )

                self.inner_function = GFPSelectIsolatedLogic(seed=seed)
            except ImportError:
                self.inner_function = instance_function_as_isolated_process(
                    name="gfp_select__isolated", seed=seed
                )
        else:
            self.inner_function = instance_function_as_isolated_process(
                name="gfp_select__isolated", seed=seed
            )

    def _black_box(self, x: np.array, context=None) -> np.ndarray:
        """
        x is string sequence which we look-up in avilable df, return median Brightness
        """
        return self.inner_function(x, context)

    def get_black_box_info(self) -> BlackBoxInformation:
        return BlackBoxInformation(
            name="gfp_select",
            max_sequence_length=237,  # max len of aaSequence
            aligned=True,  # TODO: perhaps add the fact that there is a random state here?
            fixed_length=True,
            deterministic=False,
            alphabet=AMINO_ACIDS,
            log_transform_recommended=False,
            discrete=True,
            fidelity=None,
            padding_token="",
        )


class GFPSelectionProblemFactory(AbstractProblemFactory):

    def create(
        self,
        seed: int = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        if seed is not None:
            seed_python_numpy_and_torch(seed)

        f = GFPSelectionBlackBox(
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
            seed=seed,
            force_isolation=force_isolation,
        )
        x0 = f.inner_function.x0

        problem = Problem(f, x0)

        return problem



#!/storage1/shared/software/anaconda3/bin/python3

# Copyright (C) 2020 Kristoffer Enoe Johansson <kristoffer.johansson@bio.ku.dk>

"""Module for data handling in the PRISM project

This module implments classes for parsing (PrismParser) and handling
(PrismData derived classes) of data files.

In general, calling PrismParser.read(filename) will return a data object of
the same derived class, e.g. a VariantData object.
See documentation of derived parser and data classes for help.

See README.md for general file format definitions.
"""

__version__ = 1.001

import copy
import csv
import time

import numpy as np
import pandas as pd
import yaml
from Bio import Seq, SeqIO, SeqRecord, pairwise2
from Bio.SubsMat import MatrixInfo


class PrismFormatError(Exception):
    """Exception raised for format errors in prism data files"""


class PrismValueFail(Exception):
    """Exception raised when a threshold was not met"""


class PrismParser:
    """Read and write prism data files

    Examples
    --------
    Parsing a data file
    >>> parser = PrismData.PrismParser()
    >>> data = parser.read("prism_mave/prism_mave_001_GFP.txt")

    Parsing the YAML file header only is much faster for large files
    >>> meta_data = parser.read_header("prism_mave/prism_mave_001_GFP.txt")
    """

    def read(self, filename, verbose=0):
        """Read data and meta data from a PRISM data file

        Read the meta-data section (via read_header) and data section (via Pandas.read_csv) and
        check format adherence and the consistancy between the two.

        Parameters
        ----------
        filename : string
            Name and path of file to read
        verbose : int, optional
            Level of output

        Returns
        -------
        PrismData derived instance
        """

        if verbose > 0:
            start_time = time.process_time()

        # Read meta-data and check for necessary fields
        metadata = self.read_header(filename, verbose=verbose)

        # Read data section
        dataframe = pd.read_csv(
            filename,
            delim_whitespace=True,
            comment="#",
            header=0,
            keep_default_na=True,
            na_values=["Na", "na"],
        )

        # Determine type of data
        data = None
        if dataframe.keys()[0] == "variant":
            data = VariantData(metadata, dataframe)
        # elif dataframe.keys()[0] == "resi":
        #     data = ResidueData(metadata, dataframe)
        #     checks ...
        else:
            raise PrismFormatError(
                "Could not determine data type of file %s from first column name %s"
                % (filename, dataframe.columns[0])
            )

        data.check(verbose=verbose)

        if verbose > 0:
            elapsed_time = time.process_time() - start_time
            print("Total time of parsing %s %.1f sec" % (filename, elapsed_time))

        return data

    def write(
        self, filename, data, comment_lines=None, no_overwrite_metadata=False, verbose=0
    ):
        """Write a PrismData object to a file

        Will check the consistency of variants and cloumns meta-data fields.

        Parameters
        ----------
        filename : string
            Name and path of file to write
        data : VariantData
            Data set to write
        comment_lines : list of strings
            Each list element is printed as a comment line between header and data sections
        no_overwrite_metadata : bool
            Do not allow the check function to overwrite meta data. Except instead
        """

        # Replace whitespace in column names if present
        has_whitespace = [cn.find(" ") > -0.5 for cn in data.metadata["columns"].keys()]
        if any(has_whitespace):
            if verbose > 0:
                print(
                    "White space in columns names replaced with _ (they are white-space separated)"
                )
            # Replace in data frame
            data.dataframe.columns = [
                cn.strip().replace(" ", "_") for cn in data.dataframe.columns
            ]
            # Replace in meta data
            data.metadata["columns"] = {
                k.strip().replace(" ", "_"): v
                for (k, v) in data.metadata["columns"].items()
            }

        # Check data before printing
        data.check(no_overwrite_metadata=no_overwrite_metadata, verbose=verbose)

        # if present, remove the index columns (aa_ref, etc.)
        write_cols = list(data.dataframe.columns)
        for col in data.index_column_names:
            if col in write_cols:
                write_cols.remove(col)

        # Write a file
        with open(filename, "w") as fh:
            fh.write(data.header_to_string(skip_fields=["filename"]))
            if comment_lines:
                for line in comment_lines:
                    fh.write("# " + line + "\n")
            fh.write("#\n")
            # data.dataframe.to_csv(fh, float_format='%8.4f', sep=' ', header=True, index=False, columns=write_cols,
            data.dataframe.to_csv(
                fh,
                sep=" ",
                header=True,
                index=False,
                columns=write_cols,
                quoting=csv.QUOTE_NONE,
                escapechar=" ",
                na_rep="    NA",
            )

    def is_aa_one_nat(self, sequence, additional=""):
        """Test if a sequence consists of single-letter natural amino acids only (case-insensitive)

        Parameters
        ----------
        sequence : string
            Amino acid sequence
        additional : string, optional
            Additional characters to allow, case insensitive
        """
        for aa in sequence.upper():
            if not (aa in "ACDEFGHIKLMNPQRSTVWY" or aa in additional.upper()):
                return False
        return True

    def read_header(self, filename, verbose=0):
        """Read the YAML header of a data file and run check_header

        Parameters
        ----------
        filename : string
            Path and name of the file to read
        verbose : int, optional
            Level of output

        Returns
        -------
        dict
            The meta data as a dictonary of one or more levels of dictionaries
        """
        header_begin = False
        header_str = ""
        with open(filename, "r") as file_handle:
            for line in file_handle:
                if line.replace(" ", "")[0:4] == "#---":
                    if header_begin:
                        header = yaml.safe_load(header_str)
                        header["filename"] = filename.split("/")[-1]
                        self.check_header(header)
                        return header
                    else:
                        header_begin = True
                        continue
                if header_begin:
                    if not line.strip()[0] == "#":
                        raise PrismFormatError(
                            "Missing comment char inside YAML header"
                        )
                    s = line.replace("#", "", 1)
                    header_str += s
        if header_begin:
            raise PrismFormatError("Header never ended")
        else:
            raise PrismFormatError("No header found")

    def check_header(self, header):
        """Check a header for fields required by all data files"""
        if "version" not in header.keys():
            raise PrismFormatError("Header has no 'version' field")
        if "protein" not in header.keys():
            raise PrismFormatError("Header has no 'protein' field")
        if "name" not in header["protein"].keys():
            raise PrismFormatError("Header has no 'protein: name' field")
        if "sequence" not in header["protein"].keys():
            raise PrismFormatError("Header has no 'protein: sequence' field")
        if "uniprot" not in header["protein"].keys():
            raise PrismFormatError("Header has no 'protein: uniprot' field")
        if "first_residue_number" in header["protein"].keys():
            if int(header["protein"]["first_residue_number"]) < 0:
                raise PrismFormatError("First residue number must be non-negative")
        if "columns" not in header.keys():
            raise PrismFormatError("Header has no 'columns' field")
        if "filename" in header.keys():
            data_type = header["filename"].split("_")[1]
            if data_type.lower() not in header.keys():
                raise PrismFormatError(
                    "Header has no '%s' field but filename indicates this data type"
                    % (data_type)
                )

    def write_meta_file(self, data_list, filename, output_type="csv"):
        """Write a overview file of meta-data (headers) from a list of data sets

        Parameters
        ----------
        data_list : list of PrismData-derived objects or meta data dictionaries
            The data to output
        filename : string
            Name and path of output file
        output_type : string, optional
            One of the following
            -csv: Excel-readable CSV overview of all meta-data
            -fasta: Header sequences in FASTA format
        """
        header_list = [
            data.metadata if isinstance(data, PrismData) else data for data in data_list
        ]
        if output_type == "csv":
            self.__dump_header_csv(filename, header_list)
        elif output_type == "fasta":
            self.__dump_fasta(filename, header_list)
        else:
            raise ValueError(
                "write_meta_file argument output_type must be 'csv' or 'fasta'"
            )

    def __merge_header_fields(self, header_list, mode="union"):
        """Determine a common set of header fields"""

        # Recursive grap of keys - because we can-can
        def update_keys(key_dic, dic):
            """Update key_dic with keys from dic recursively"""
            for key in dic.keys():
                if key not in key_dic.keys():
                    key_dic[key] = {}
                if isinstance(dic[key], dict):
                    update_keys(key_dic[key], dic[key])

        # Read all header keys
        common_fields = {}
        for header in header_list:
            update_keys(common_fields, header)
        # Return according to mode
        if mode.lower() == "union":
            return common_fields
        elif mode.lower() == "intersect":
            raise NotImplementedError(
                "__merge_header_fields mode==intersect not implemented"
            )
        else:
            raise ValueError("Header merge mode must be 'union' or 'intersect'")

    def __dump_header_csv(self, filename, header_list):
        common_header_keys = self.__merge_header_fields(header_list, mode="union")

        # Columns fields have different names so rename to column01, column02, etc
        ncol_list = [len(header["columns"].keys()) for header in header_list]
        new_col_names = ["column%02d" % (c + 1) for c in range(max(ncol_list))]
        common_header_keys["columns"] = dict(
            zip(new_col_names, [{}] * len(new_col_names))
        )

        with open(filename, "w", newline="") as csvfile:
            csv_dialect = csv.excel()
            csv_dialect.delimiter = ";"
            writer = csv.writer(csvfile, dialect=csv_dialect)
            # First layer of keys
            row = []
            for key in common_header_keys:
                row += [key] + [""] * np.max([0, len(common_header_keys[key]) - 1])
            writer.writerow(row)
            # Second layer of keys
            row = []
            for key in common_header_keys:
                if len(common_header_keys[key]) == 0:
                    row += [key]
                else:
                    row += [kkey for kkey in common_header_keys[key].keys()]
            writer.writerow(row)
            # Fill in values for each header
            for header in header_list:
                row = []
                for key in common_header_keys.keys():
                    if key not in header:
                        row += [""] * np.max([1, len(common_header_keys[key])])
                    elif len(common_header_keys[key].keys()) == 0:
                        row += [header[key]]
                    elif key == "columns":
                        for keyi in range(len(common_header_keys[key])):
                            key_list = list(header["columns"].keys())
                            value_list = list(header["columns"].values())
                            if keyi < len(header["columns"]):
                                row += [key_list[keyi] + ": " + value_list[keyi]]
                            else:
                                row += [""]
                    else:
                        for kkey in common_header_keys[key].keys():
                            if kkey in header[key].keys():
                                row += [header[key][kkey]]
                            else:
                                row += [""]
                writer.writerow(row)

    def __dump_fasta(self, filename, header_list):
        sequences = []
        for header in header_list:
            file_id = (
                header["filename"].rsplit(".", 1)[0] + "_v" + str(header["version"])
            )
            seq = Seq.Seq(header["protein"]["sequence"])
            sequences.append(SeqRecord.SeqRecord(seq, id=file_id, description=""))
        with open(filename, "w", newline="") as file_handle:
            count = SeqIO.write(sequences, file_handle, "fasta")
        return count


class PrismData:
    """Base class of PRISM data types

    This class is thought as abstract and should not be instantiated. See documentation of derived classes.
    """

    def __init__(self, metadata, dataframe):
        """Constructor

        See documentation of derived classes
        """
        self.metadata = metadata
        self.dataframe = dataframe
        self.index_column_names = None

    def __repr__(self):
        """Representation - used e.g. for interactive printing"""
        return self.header_to_string() + str(self.dataframe)

    def __str__(self):
        """Cast to string - used e.g. by print"""
        return "%s object of protein %s (uniprot %s) with dataframe shape %s" % (
            type(self),
            self.metadata["protein"]["name"],
            self.metadata["protein"]["uniprot"],
            str(self.dataframe.shape),
        )

    def copy(self):
        """Make a deep copy of self"""
        new_meta = copy.deepcopy(self.metadata)
        new_dataframe = self.dataframe.copy(deep=True)
        new_obj = PrismData(new_meta, new_dataframe)
        new_obj.__class__ = self.__class__
        return new_obj

    def seq_from_data(self):
        """Builds a sequence based on aa_ref and resi data columns

        Data is assumed to have index columns resi and aa_ref
        """

        if "aa_ref" not in self.dataframe.columns:
            self.add_index_columns()

        n_res = 0
        seq = np.full(100, "X")
        for aa_list, resi_list in zip(self.dataframe["aa_ref"], self.dataframe["resi"]):
            for aa, resi in zip(aa_list, resi_list):
                n_res = max(n_res, resi)
                if n_res > len(seq):
                    seq = np.concatenate(
                        [seq, np.full(max(100, n_res - len(seq)), "X")]
                    )
                if seq[resi - 1] == "X":
                    seq[resi - 1] = aa
                elif seq[resi - 1] != aa:
                    raise PrismFormatError(
                        "Data reference amino acid mismatch at position %d: % and %s"
                        % (resi, seq[resi - 1], aa)
                    )
        return "".join(seq[:n_res])

    def check_header_and_data_sequence(self, verbose=0):
        """Check reference sequence from header against the data"""
        data_seq = self.seq_from_data()
        meta_seq = self.metadata["protein"]["sequence"]
        first_resn = (
            int(self.metadata["protein"]["first_residue_number"])
            if "first_residue_number" in self.metadata["protein"].keys()
            else 1
        )
        meta_seq = (
            "X" * (first_resn - 1) + meta_seq
        )  # For checking purposes, matching indices makes things easier

        # Check sequence length
        if len(data_seq) < len(meta_seq):
            # Add X for unobserved positions
            data_seq = data_seq + "X" * (len(meta_seq) - len(data_seq))
        elif len(data_seq) != len(meta_seq):
            raise PrismFormatError("Data has more residues than header sequence")

        # Check all substitution amino acids against header sequence
        match = [
            aa1 == aa2 if aa1 != "X" and aa2 != "X" else True
            for (aa1, aa2) in zip(meta_seq, data_seq)
        ]
        if not np.all(match):
            mismatch = np.where(np.logical_not(match))[0] + 1
            raise PrismFormatError(
                "Mismatch between header sequence and data sequence at positions %s:\n%s\n%s"
                % (str(mismatch), meta_seq, data_seq)
            )
        if verbose > 0:
            print(
                "Data and header amino acid sequence match with %d of %d positions not observed in data"
                % (data_seq.upper().count("X"), len(meta_seq))
            )
        return data_seq

    def header_to_string(self, skip_fields=[]):
        """Format meta-data into a string"""

        # Recursive function to write the next level of a dictionary
        def dump_subfields(dic, indent_level):
            return_str = ""
            dict_keys = list(dic.keys())
            # For the top level, order fields (dict is not ordered)
            if indent_level == 0:
                dict_keys_first = list()
                dict_keys_last = list()
                for k in ["version", "protein"]:
                    if k in dict_keys:
                        dict_keys.remove(k)
                        dict_keys_first.append(k)
                for k in ["variants", "columns", "filename"]:
                    if k in dict_keys:
                        dict_keys.remove(k)
                        dict_keys_last.append(k)
                dict_keys = dict_keys_first + dict_keys + dict_keys_last
            # Make new indent and run recursively or dump values
            for key in dict_keys:
                if key in skip_fields:
                    continue
                if isinstance(dic[key], dict):
                    return_str += "# " + "    " * indent_level + key + ":\n"
                    return_str += dump_subfields(dic[key], indent_level + 1)
                else:
                    # format floats values and use standard cast for others
                    s = (
                        "%.3f" % (dic[key])
                        if isinstance(dic[key], float)
                        else str(str(dic[key]))
                    )
                    return_str += "# " + "    " * indent_level + key + ": " + s + "\n"
            return return_str

        return_str = ""
        return_str += "# --------------------\n"
        return_str += dump_subfields(self.metadata, 0)
        return_str += "# --------------------\n"
        return_str += "#\n"
        return return_str

    def check_column_names(self, verbose=0):
        # Check column names - case sensitive
        meta_colnames = list(self.metadata["columns"].keys()) + [
            "n_mut",
            "aa_ref",
            "resi",
            "aa_var",
        ]
        data_colnames = self.dataframe.keys()[
            1:
        ]  # first column is data specific and not in header
        for cn in data_colnames:
            if cn not in meta_colnames:
                raise PrismFormatError(
                    "Could not find column name '%s' in header" % (cn)
                )
        for cn in meta_colnames:
            if cn not in data_colnames:
                raise PrismFormatError(
                    "Could not find header column name '%s' in data" % (cn)
                )
        if verbose > 0:
            print("Column names checks ok: %s" % (",".join(self.dataframe.keys())))


class VariantData(PrismData):
    """Container class for protein variant data

    Parameters
    ----------
    metadata : dict
        Metadata for data set
    dataframe : pandas.DataFrame
        Data

    Examples
    --------
    Get meta-data dictionary
    >>> meta_data = VariantData.metadata

    Get sequence from header
    >>> sequence = meta_data['protein']['sequence']

    Get pandas.DataFrame
    >>> data_frame = VariantData.dataframe

    Get all double mutants
    >>> data_frame[data_frame['n_mut']==2]

    Get all variants with substitutions into Ala
    >>> data_frame = VariantData.get_var_into_aa('A')

    Get all variants involving positions 30-45 (both included)
    >>> data_frame = VariantData.get_var_at_pos(range(30,45+1))

    Get a complete site-saturation (single mutant) library
    >>> data_frame = VariantData.saturate_variants()

    Merge three VariantData files and retrieve the union of variants (merge=outer)
    >>> merged_data = data1.merge([data2, data3], merge="outer")
    """

    # To be implmented:
    # Get all (multi-mutant) variants that involves D83V and L11P
    # >>> subst2var = VariantData.make_subst2var_map()
    # >>> data_frame = VariantData.dataframe[subst2var['D83V'] and subst2var['L11P']]

    def add_index_columns(self):
        var_split = [
            var.split(":") if var != "WT" else [] for var in self.dataframe["variant"]
        ]
        resi_ll = [
            [int(sub[1:-1]) for sub in var] if len(var) > 0 else [] for var in var_split
        ]
        aa_ref_ll = [
            [mut[0] for mut in var] if len(var) > 0 else [] for var in var_split
        ]
        aa_mut_ll = [
            [mut[-1] for mut in var] if len(var) > 0 else [] for var in var_split
        ]
        self.dataframe["n_mut"] = [len(var) for var in var_split]
        self.dataframe["aa_ref"] = aa_ref_ll
        self.dataframe["resi"] = resi_ll
        self.dataframe["aa_var"] = aa_mut_ll
        self.index_column_names = ["n_mut", "aa_ref", "resi", "aa_var"]

    def remove_index_columns(self):
        """Remove the indexing columns, i.e. ['n_mut','aa_ref','resi','aa_var']"""
        if "aa_ref" in self.dataframe.columns:
            self.dataframe.drop(
                ["n_mut", "aa_ref", "resi", "aa_var"], axis=1, inplace=True
            )
        self.index_column_names = None

    def get_var_into_aa(self, aa, multimutant_mode="any"):
        """Get a data frame of variants with substitutions into a given amino acid

        Parameters
        ----------
        aa : character
            Amino acid in one-letter code
        multimutant_mode : string
            Mode of action for multi-mutants
            - any : At least one substitution per variant into aa required
            - all : All substitutions in variants must be into aa
            - exclude : Only return single mutants
        """
        # Check argument
        if multimutant_mode not in ["any", "all", "exclude"]:
            raise ValueError(
                "Function get_var_into_aa argument multimutant_mode must be 'any', 'all' or 'exclude'"
            )
        # Make a series of boleans to select rows
        if multimutant_mode == "any":
            bool_mask = [aa in aa_list for aa_list in self.dataframe["aa_var"]]
        elif multimutant_mode == "all":
            bool_mask = [
                np.all(np.array(aa_list) == aa) if len(aa_list) > 0 else False
                for aa_list in self.dataframe["aa_var"]
            ]
        elif multimutant_mode == "exclude":
            bool_mask = [aa_list == [aa] for aa_list in self.dataframe["aa_var"]]
        else:
            assert False
        return self.dataframe[bool_mask]

    def get_var_from_aa(self, aa, multimutant_mode="any"):
        # Exclude synonymous and nonsense?
        """Get a data frame of variants with substitutions from a given amino acid

        Parameters
        ----------
        aa : character
            Amino acid in one-letter code
        multimutant_mode : string
            Mode of action for multi-mutants
            - any : At least one substitution per variant from aa required
            - all : All substitutions in variants must be from aa
            - exclude : Only return single mutants
        """
        # Check argument
        if multimutant_mode not in ["any", "all", "exclude"]:
            raise ValueError(
                "Function get_var_from_aa argument multimutant_mode must be 'any', 'all' or 'exclude'"
            )
        # Make a series of boleans to select rows
        if multimutant_mode == "any":
            bool_mask = [aa in aa_list for aa_list in self.dataframe["aa_ref"]]
        elif multimutant_mode == "all":
            bool_mask = [
                np.all(np.array(aa_list) == aa) if len(aa_list) > 0 else False
                for aa_list in self.dataframe["aa_ref"]
            ]
        elif multimutant_mode == "exclude":
            bool_mask = [aa_list == [aa] for aa_list in self.dataframe["aa_ref"]]
        else:
            assert False
        return self.dataframe[bool_mask]

    def get_var_at_pos(self, target_resi, mode="any"):
        # Exclude synonymous and nonsense?
        """Get a data frame of variants with substitutions in one or more positions

        Parameters
        ----------
        target_resi : int or iterable of int
            Residue number as given in variants
        mode : string
            Mode of action for multi-mutants
            - any : At least one substitution in one of the given position
            - all : Substitutions at all given position
            - exact : Substitutions at all given position and no others
        """
        # Check argument mode
        if mode not in ["any", "all", "exact"]:
            raise ValueError(
                "Function get_var_from_aa argument mode must be 'any', 'all' or 'exact'"
            )
        # Check argument target_resi
        if isinstance(target_resi, int):
            target_resi = [target_resi]
        elif not isinstance(target_resi, list):
            target_resi = list(target_resi)
        target_resi = list(np.sort(target_resi))
        if mode == "any":
            bool_mask = [
                (
                    np.any([ti in resi_list for ti in target_resi])
                    if len(resi_list) > 0
                    else False
                )
                for resi_list in self.dataframe["resi"]
            ]
        elif mode == "all":
            bool_mask = [
                (
                    np.all([ti in resi_list for ti in target_resi])
                    if len(resi_list) > 0
                    else False
                )
                for resi_list in self.dataframe["resi"]
            ]
        elif mode == "exact":
            bool_mask = [
                resi_list == target_resi for resi_list in self.dataframe["resi"]
            ]
        else:
            assert False
        return self.dataframe[bool_mask]

    def get_subst(self, subst, single_only=True):
        """Get a data frame of variants that all contain a given substitution

        Parameters
        ----------
        subst : string
            Substitution, e.g. M1A
        single_only : bool
            Only return single mutants. If false, all multi-mutants with the given substitution will be returned
        """
        bool_mask = [
            subst in var.split(":") if var != "WT" else False
            for var in self.dataframe["variant"]
        ]
        # Add requirement on number of mutations if requested
        if single_only:
            bool_mask = [
                a and b for a, b in zip(bool_mask, self.dataframe["n_mut"] == 1)
            ]
        # Return data frame slice
        return self.dataframe[bool_mask]

    def order_variants(self, aa_order="CDEKRHNQAGSTVMLIFYWP=*~"):
        """Order variant according to number of substitutions, residue number, and a specified order of
        amino acids. Non-specified amino acids are keps in original order. Inserts are not supported.
        """
        n_aa = len(aa_order)
        if len(set(aa_order)) != n_aa:
            raise ValueError(
                "Argument aa_order to VariantData.order_variants must be unique characters"
            )
        aa_dict = dict(zip(aa_order, range(n_aa)))
        aa_dict["WT"] = -1
        aa_list = [l[0] if len(l) > 0 else "WT" for l in self.dataframe["aa_var"]]
        self.dataframe["aa0"] = [
            aa_dict[aa] if aa in aa_dict.keys() else n_aa for aa in aa_list
        ]
        self.dataframe["resi0"] = [
            l[0] if len(l) > 0 else 0 for l in self.dataframe["resi"]
        ]
        self.dataframe.sort_values(by=["n_mut", "resi0", "aa0"], inplace=True)
        self.dataframe.index = range(self.dataframe.shape[0])
        self.dataframe.drop(columns=["aa0", "resi0"], inplace=True)

    def saturate_variants(self, aa_order="CDEKRHNQAGSTVMLIFYWP", verbose=0):
        """Saturate variant list to make a full site-saturation library

        Order single mutants according to position and target amino acid and add NA
        rows for missing single variants. Inserts are not supported.

        Parameters
        ----------
        aa_order : string
            Which amino acids to include and the order, default CDEKRHNQAGSTVMLIFYWP
        verbose : int, optional
            Level of output
        """
        n_aa = len(aa_order)
        if len(set(aa_order)) != n_aa:
            raise ValueError(
                "Argument aa_order to VariantData.saturate_variants must be unique characters"
            )
        seq = self.metadata["protein"]["sequence"]
        first_resn = (
            int(self.metadata["protein"]["first_residue_number"])
            if "first_residue_number" in self.metadata["protein"].keys()
            else 1
        )
        var_list = [
            (
                seq[resi] + str(resi + first_resn) + var
                if seq[resi] != var
                else seq[resi] + str(resi + first_resn) + "="
            )
            for resi in range(len(seq))
            for var in aa_order
        ]
        if verbose > 0:
            remove_set = set(self.dataframe["variant"]) - set(var_list)
            if len(remove_set) > 0:
                if len(remove_set) > 500:
                    rm_list = list(remove_set)
                    print(
                        "Saturate variants will remove the following %d variants:\n%s\n... [skip %d variants] ...\n%s"
                        % (
                            len(rm_list),
                            ",".join([str(i) for i in rm_list[0:25]]),
                            len(rm_list) - 50,
                            ",".join([str(i) for i in rm_list[-25:]]),
                        )
                    )
                else:
                    print(
                        "Saturate variants will remove the following %d variants:\n%s"
                        % (len(remove_set), ",".join([str(i) for i in remove_set]))
                    )
        new_dataframe = pd.DataFrame({"variant": var_list})
        new_dataframe = new_dataframe.merge(
            self.dataframe, on="variant", how="left", copy=False
        )
        self.dataframe = new_dataframe
        self.add_index_columns()
        self.metadata["variants"] = self.calc_variants_metadata(verbose=verbose)

    def __missing_mask(self, verbose=0):
        """Private helper function for clean_na_variants and others

        Returns a bool array that marks rows with NA only. Should be fast."""
        # pd.DataFrame.iterrows() is not vectorized and very slow, DataFrame.apply is not too good either
        # return([np.all(row[1]) for row in self.dataframe.drop(columns=['variant','n_mut','aa_ref','resi','aa_var']).isnull().iterrows()])
        if "aa_ref" in self.dataframe.columns:
            return np.all(
                self.dataframe.drop(
                    columns=["variant", "n_mut", "aa_ref", "resi", "aa_var"]
                ).isnull(),
                axis=1,
            )
        else:
            return np.all(self.dataframe.drop(columns="variant").isnull(), axis=1)

    def clean_na_variants(self, verbose=0):
        """Remove variants rows with all data missing"""
        missing_mask = self.__missing_mask()
        i = np.where(missing_mask)[0]
        if len(i) > 0:
            self.dataframe.drop(self.dataframe.index[i], axis=0, inplace=True)
            self.dataframe.index = range(self.dataframe.shape[0])
            if verbose > 0:
                print("Removed %d variant rows with all data missing" % (len(i)))

    def check_variants_metadata(self, no_overwrite=False, verbose=0):
        recalc_variants = self.calc_variants_metadata(verbose=verbose)
        if "variants" in self.metadata:
            # Number of variants
            if "number" in self.metadata["variants"].keys():
                if recalc_variants["number"] != self.metadata["variants"]["number"]:
                    s = (
                        "Header field 'variants:number' is %d but %d variants (possible including WT) were found"
                        % (
                            self.metadata["variants"]["number"],
                            recalc_variants["number"],
                        )
                    )
                    if no_overwrite:
                        raise PrismFormatError(s)
                    elif verbose > 0:
                        print("WARNING: " + s)
            elif verbose > 0 and not no_overwrite:
                print(
                    "Will add variants:number: %d to header"
                    % (recalc_variants["number"])
                )

            # Coverage
            if "coverage" in self.metadata["variants"].keys():
                if (
                    np.abs(
                        recalc_variants["coverage"]
                        - self.metadata["variants"]["coverage"]
                    )
                    > 0.01
                ):
                    s = (
                        "Header field 'variants:coverage' is %.2f but variants calculated to cover %.2f"
                        % (
                            self.metadata["variants"]["coverage"],
                            recalc_variants["coverage"],
                        )
                    )
                    if no_overwrite:
                        raise PrismFormatError(s)
                    elif verbose > 0:
                        print("WARNING: " + s)
            elif verbose > 0 and not no_overwrite:
                print(
                    "Will add variants:coverage: %.2f to header"
                    % (recalc_variants["coverage"])
                )

            # Depth
            if "depth" in self.metadata["variants"].keys():
                if (
                    np.abs(
                        recalc_variants["depth"] - self.metadata["variants"]["depth"]
                    )
                    > 0.01
                ):
                    s = (
                        "Header field 'variants:depth' is %.2f but variants calculated to depth %.2f"
                        % (self.metadata["variants"]["depth"], recalc_variants["depth"])
                    )
                    if no_overwrite:
                        raise PrismFormatError(s)
                    elif verbose > 0:
                        print("WARNING: " + s)
            elif verbose > 0 and not no_overwrite:
                print(
                    "Will add variants:depth: %.2f to header"
                    % (recalc_variants["depth"])
                )

            # Helper function
            def strip_all(s):
                return (
                    s.lower()
                    .replace(".", "")
                    .replace("-", "")
                    .replace("+", "")
                    .replace("/", "")
                    .replace(" ", "")
                )

            # Variant width
            if "width" in self.metadata["variants"].keys():
                if recalc_variants["width"] == "single mutants":
                    if strip_all(self.metadata["variants"]["width"]) not in [
                        "singlemutants",
                        "singlemutant",
                        "singlemut",
                        "single",
                        "singleonly",
                    ]:
                        s = (
                            "Header field 'variants:width' is '%s' but only single mutants were found - use 'single mutants'"
                            % (self.metadata["variants"]["width"])
                        )
                        if no_overwrite:
                            raise PrismFormatError(s)
                        elif verbose > 0:
                            print("WARNING: " + s)
                elif recalc_variants["width"] == "multi mutants":
                    if strip_all(self.metadata["variants"]["width"]) not in [
                        "multimutants",
                        "multimutant",
                        "multimut",
                        "multi",
                    ]:
                        # multimut = [var_list[i] for i in np.where(n_mut > 2)[0]]
                        s = (
                            "Header field 'variants:width' is '%s' but the following multi-mutant were found - use 'multi mutants'"
                            % (self.metadata["variants"]["width"])
                        )
                        if no_overwrite:
                            raise PrismFormatError(s)
                        elif verbose > 0:
                            print("WARNING: " + s)
                elif recalc_variants["width"] == "single and double mutants":
                    if strip_all(self.metadata["variants"]["width"]) not in [
                        "singleanddoublemutants",
                        "singleanddoublemutant",
                        "singleanddouble",
                    ]:
                        s = (
                            "Header field 'variants:width' is '%s' but variants are single and double mutants - use 'single and double mutants'"
                            % (self.metadata["variants"]["width"])
                        )
                        if no_overwrite:
                            raise PrismFormatError(s)
                        elif verbose > 0:
                            print("WARNING: " + s)
                else:
                    raise ValueError(
                        "Recalculated variants width has an unknown value '%s'"
                        % (recalc_variants["width"])
                    )
            elif verbose > 0 and not no_overwrite:
                print(
                    "Will add variants:width: %s to header" % (recalc_variants["width"])
                )
        else:
            if no_overwrite:
                raise PrismFormatError("No variants field in meta data")

        if not no_overwrite:
            self.metadata["variants"] = recalc_variants

    def calc_variants_metadata(self, verbose=0, return_missing_mask=False):
        ret = {}

        if verbose > 1:
            print("===============================================================")
            print("====    Calculate variants meta-data fields with timing    ====")
            start_time = time.process_time()
            print(
                "Calc number of variants (incl. missing_mask) ...", end=" ", flush=True
            )

        # Number of variants
        missing_mask = self.__missing_mask(verbose=verbose)
        ret["number"] = len(self.dataframe["variant"]) - np.sum(missing_mask)

        if verbose > 1:
            elapsed_time = time.process_time() - start_time
            print("done. Elapsed time %.1f sec" % (elapsed_time))
            start_time = time.process_time()
            print("Calc coverage and depth of variants ...", end=" ", flush=True)

        # Coverage and depth
        # Positions with substitutions
        non_missing_indices = np.where([not b for b in missing_mask])[0]
        flat_resi_list = [
            resi
            for resi_list in self.dataframe["resi"][non_missing_indices]
            for resi in resi_list
        ]
        (resi_list, resi_counts) = np.unique(flat_resi_list, return_counts=True)

        # Coverage is positions with substitutions per residue in reference sequence
        ret["coverage"] = len(resi_list) / len(self.metadata["protein"]["sequence"])
        # Depth is the average number of substitutions per position with substitutions
        ret["depth"] = np.mean(resi_counts)

        if verbose > 1:
            elapsed_time = time.process_time() - start_time
            print("done. Elapsed time %.1f sec" % (elapsed_time))
            start_time = time.process_time()
            print("Calc width of variants ...", end=" ", flush=True)

        # Width
        n_mut = self.dataframe["n_mut"]
        if np.all(n_mut <= 1):
            ret["width"] = "single mutants"
        elif np.max(n_mut) > 2:
            ret["width"] = "multi mutants"
        else:
            ret["width"] = "single and double mutants"

        if verbose > 1:
            elapsed_time = time.process_time() - start_time
            print("done. Elapsed time %.1f sec" % (elapsed_time))
            print("===============================================================")

        if return_missing_mask:
            return (ret, missing_mask)
        else:
            return ret

    def check(self, no_overwrite_metadata=False, verbose=0):
        """Perform all checks of data and meta data

        Checks formatting of all variants in the first column and sorts multimutants
        and that the list is unique. Adds index columns.

        Consistancy checks of data and meta data includes column names, reference amino
        acid sequence and all variants sub fields.

        Parameters
        ----------
        no_overwrite_metadata : bool
            Allow overwriting of the variants meta-data fields
        verbose : int, optional
            Level of output
        """
        # Check and update variant column
        if verbose > 1:
            print("===============================================================")
            print("====     Check data section variant list with timing       ====")
            start_time = time.process_time()
            print("Make var_list ...", end=" ", flush=True)

        var_list = [s.upper() for s in self.dataframe["variant"]]

        # At least 2 characters for 'WT'
        if np.any([len(var) < 2 for var in var_list]):
            il = np.where([len(var) for var in var_list] < 2)[0]
            raise PrismFormatError(
                "Bad variants: %s" % (str([var_list[i] for i in il]))
            )

        if verbose > 1:
            elapsed_time = time.process_time() - start_time
            print("done. Elapsed time %.1f sec" % (elapsed_time))
            start_time = time.process_time()
            print("Sort variants ...", end=" ", flush=True)

        # Sort multi-mutants according to residue number
        var_split_unordered = [
            var.split(":") if var != "WT" else [] for var in var_list
        ]
        resi_ll_unordered = [
            [int(subst[1:-1]) for subst in var] if len(var) > 0 else []
            for var in var_split_unordered
        ]

        # Make a new variant list with ordered multi-mutants
        var_split = []
        for ivar in range(len(var_split_unordered)):
            if len(set(resi_ll_unordered[ivar])) < len(resi_ll_unordered[ivar]):
                raise PrismFormatError(
                    "Variant %d %s has more substitutions on the same position"
                    % (ivar + 1, var_list[ivar])
                )
            var_split.append(
                [
                    var_split_unordered[ivar][i]
                    for i in np.array(resi_ll_unordered[ivar]).argsort()
                ]
            )

        if verbose > 1:
            elapsed_time = time.process_time() - start_time
            print("done. Elapsed time %.1f sec" % (elapsed_time))
            start_time = time.process_time()
            print("Split all substitutions ...", end=" ", flush=True)

        # Split substitutions in amino acids and residue number
        resi_ll = [
            [int(sub[1:-1]) for sub in var] if len(var) > 0 else [] for var in var_split
        ]
        aa_ref_ll = [
            [mut[0] for mut in var] if len(var) > 0 else [] for var in var_split
        ]
        aa_mut_ll = [
            [mut[-1] for mut in var] if len(var) > 0 else [] for var in var_split
        ]

        if verbose > 1:
            elapsed_time = time.process_time() - start_time
            print("done. Elapsed time %.1f sec" % (elapsed_time))
            start_time = time.process_time()
            print("Make new_var_list ...", end=" ", flush=True)

        # Check all substitutions and make new variant list
        new_var_list = []
        for ivar in range(len(resi_ll)):
            subst_list = []
            for isub in range(len(resi_ll[ivar])):
                aa1 = aa_ref_ll[ivar][isub]
                resi = resi_ll[ivar][isub]
                aa2 = aa_mut_ll[ivar][isub]
                if not PrismParser.is_aa_one_nat(None, aa2, additional="*=~"):
                    raise PrismFormatError(
                        "Substitution %s%d%s is not to a standard amino acid"
                        % (aa1, resi, aa2)
                    )
                if aa1 == aa2:
                    if verbose > 0:
                        print(
                            "Synonymous substitution %s%d%s renamed to %s%d="
                            % (aa1, resi, aa2, aa1, resi)
                        )
                    aa_mut_ll[ivar][isub] = "="
                subst_list.append(
                    "%s%d%s"
                    % (
                        aa_ref_ll[ivar][isub],
                        resi_ll[ivar][isub],
                        aa_mut_ll[ivar][isub],
                    )
                )
            new_var_list.append(":".join(subst_list) if len(subst_list) > 0 else "WT")

        # Check if variants are unique
        (var_unique, var_counts) = np.unique(new_var_list, return_counts=True)
        if np.any(var_counts > 1):
            raise PrismFormatError(
                "The following variants are listed multiple times:\n%s"
                % (var_unique[np.where(var_counts > 1)[0]])
            )

        # Update data frame
        self.dataframe["variant"] = new_var_list

        # Update index columns now that we have them
        self.dataframe["n_mut"] = [len(var) for var in var_split]
        self.dataframe["aa_ref"] = aa_ref_ll
        self.dataframe["resi"] = resi_ll
        self.dataframe["aa_var"] = aa_mut_ll
        self.index_column_names = ["n_mut", "aa_ref", "resi", "aa_var"]

        if verbose > 1:
            elapsed_time = time.process_time() - start_time
            print("done. Elapsed time %.1f sec" % (elapsed_time))
            print("===============================================================")

        self.check_column_names(verbose=verbose)
        self.check_header_and_data_sequence(verbose=verbose)
        self.check_variants_metadata(
            no_overwrite=no_overwrite_metadata, verbose=verbose
        )

    def to_new_reference(
        self,
        target_seq=None,
        first_resn=None,
        verbose=0,
        min_identity=0.8,
        min_coverage=0.1,
        mismatch="remove",
        allow_inserts=True,
        allow_deletions=True,
    ):
        """Map variants from data file onto a new reference sequence

        This is the workhorse of data set merging

        Parameters
        ----------
        target_seq : string
            Amino acid sequence that variants are aligned to
        first_resn: integer >= 0
            If None, the header value is used if present, else one. If given the header value is replaced
        min_identity : float in [0.0, 1.0], optional
            Raise exception if the sequence identity is less than this (in covered region)
        min_coverage : float in [0.0, 1.0], optional
            Raise exception if the fraction of matched positions (i.e. not indel/gap) is less than this
            This is a symmetric kind of coverage: (N_res-N_indel)/N_res where N_res is the lengths of the
            longest sequence
        mismatch : string, optional
            How to handle mismatches in the alignment
            -convert: Change reference amino acid to match the new reference
            -remove: Remove variants that involve mismatch positions
            -not_allowed: Raise exception is a mismatch is incountered
        allow_inserts : bool, optional
            How to handle target sequence position missing in the sequence of the data
            If true, the residue numbering will be ajusted accordingly
        allow_deletions : bool, optional
            How to handle data sequence positions missing in the target sequence
            If true, the residue numbering will be ajusted accordingly
        verbose : int, optional
            Level of output
        """

        # Determine new sequence offset and how much to shift residue numbering
        first_resn_meta = (
            int(self.metadata["protein"]["first_residue_number"])
            if "first_residue_number" in self.metadata["protein"].keys()
            else 1
        )
        if target_seq is None and first_resn is None:
            # Nothing to do
            return self
        elif first_resn is None:
            # Only alignment to target seq
            first_resn = first_resn_meta
            resi_shift_init = 0
        else:
            # Shift residue numbering (and possible align to target sequence)
            resi_shift_init = first_resn - first_resn_meta

        if first_resn < 0:
            raise ValueError(
                "Argument first_resn to VariantData.to_new_reference must be a non-negative integer"
            )

        seq = self.metadata["protein"]["sequence"]
        n_res_data = len(seq)

        # Which original data residue indices (0-based) to change
        resi_rm = []
        resi_shift = np.full(n_res_data, resi_shift_init)
        aa_change = {}
        if target_seq is not None:
            if not PrismParser.is_aa_one_nat(None, target_seq, "X"):
                raise ValueError(
                    "Argument target_seq to VariantData.to_new_reference must be a single-letter amino acid string (or None)"
                )
            n_res_target = len(target_seq)
            # Align sequences
            # MatrixInfo supports a wildcard 'X' so use upper case. Open and extend penalty 11 and 1 is blast default
            # 2DO: Consider to write gap_function where X--XX is 2 openings and --XXX is only one, i.e. NT-gap is not an opening
            align = pairwise2.align.globalds(
                target_seq.upper(), seq.upper(), MatrixInfo.blosum62, -3, -1
            )
            # Check for alternative alignments
            if len(align) > 1 and verbose > 0:
                print("Found %d alignments" % (len(align)))
            if verbose > 0:
                # Dump verbose number of alignments
                for ia in range(min([verbose, len(align)])):
                    print("==== alignment %d of %d ====" % (ia + 1, len(align)))
                    print(pairwise2.format_alignment(*align[ia]))
            # Check for equally good alignments, a good alignment has high score
            if len(align) > 1:
                align_scores = np.array([a[2] for a in align])
                if any(align_scores[0] - align_scores[1:] < 1e-3):
                    print(
                        "WARNING: There are alignments for %s with same or better scores (try verbose). Sores: %s"
                        % (self.metadata["protein"]["name"], str(align_scores))
                    )
            # Use first alignment
            align = align[0]
            assert len(align[0]) == len(align[1])
            n_align = len(align[0])
            # Number to convert 0-based alignment index to 0-based data (original) residue numbering
            ialign_shift = 0
            n_indel = n_match = n_mismatch = 0
            # Iterate over alignment, possible including gaps
            for ialign in range(len(align[0])):
                if align[0][ialign] == "-":
                    # Deletion: Position in data sequence not present in target sequence
                    if not allow_deletions:
                        raise PrismValueFail(
                            "Aborting alignment due to deletion at position %d which is not allowed"
                            % (ialign + ialign_shift)
                        )
                    # Remove all substitutions at this position
                    resi_rm.append(ialign + ialign_shift)
                    # Downstream position are shifted to the left
                    resi_shift[(ialign + ialign_shift) :] -= 1
                    n_indel += 1
                elif align[1][ialign] == "-":
                    # Insertion: Position in target sequence not present in data sequence
                    if not allow_inserts:
                        raise PrismValueFail(
                            "Aborting alignment due to insertion at position %d which is not allowed"
                            % (ialign + ialign_shift)
                        )
                    resi_shift[(ialign + ialign_shift) :] += 1
                    # Shift conversion number
                    ialign_shift -= 1
                    n_indel += 1
                elif align[0][ialign] != align[1][ialign]:
                    # Mismatch of 'X' always allowed since it is always mismatched
                    # Variants matched to X are always removed because position is ambiguous if adjacent to gap and because X is used for numbering offset
                    if align[0][ialign] == "X" or align[1][ialign] == "X":
                        resi_rm.append(ialign + ialign_shift)
                    elif mismatch == "not_allowed":
                        raise PrismValueFail(
                            "Aborting alignment due to mismatch at position %d which is not allowed"
                            % (ialign + ialign_shift)
                        )
                    elif mismatch == "remove":
                        # Mark the substitutions at this position for removal
                        resi_rm.append(ialign + ialign_shift)
                    elif mismatch == "convert":
                        # Change reference amino acid
                        aa_change[ialign + ialign_shift] = align[0][ialign]
                    else:
                        raise ValueError(
                            "mismatch argument must be in [not_allowed,convert,remove]"
                        )
                    n_mismatch += 1
                else:
                    n_match += 1
            # Is alignment ok?
            n_res = max(n_res_target, n_res_data)
            coverage = (n_res - n_indel) / n_res
            if coverage < min_coverage:
                raise PrismValueFail(
                    "Sequence coverage %.3f (%d/%d) below threshold of %.3f"
                    % (coverage, n_res - n_indel, n_res, min_coverage)
                )
            identity = (n_res_target - n_mismatch) / n_res_target
            if identity < min_identity:
                raise PrismValueFail(
                    "Sequence identity %.3f below threshold of %.3f"
                    % (identity, min_identity)
                )
            # Update header if alignment is ok
            self.metadata["protein"]["sequence"] = target_seq
            # Report actions
            if verbose > 0:
                print(
                    "Making new reference of %s with coverage %.3f and identity %.3f. Maximum shift in residue number is %d"
                    % (
                        self.metadata["filename"],
                        coverage,
                        identity,
                        max(abs(resi_shift)),
                    )
                )
            if verbose > 0 and len(resi_rm) > 0:
                print(
                    "New reference sequence will remove variants at %d positions (mismatch setting: %s; indels: %d):\n%s"
                    % (
                        len(resi_rm),
                        mismatch,
                        n_indel,
                        ",".join([str(resi + first_resn_meta) for resi in resi_rm]),
                    )
                )
            if verbose > 0 and len(aa_change) > 0:
                print(
                    "New reference sequence will change variants at %d positions (mismatch setting: %s; indels: %d):\n%s"
                    % (
                        len(aa_change),
                        mismatch,
                        n_indel,
                        ",".join(
                            [
                                str(int(resi) + first_resn_meta)
                                for resi in aa_change.keys()
                            ]
                        ),
                    )
                )

        if first_resn is not None:
            self.metadata["protein"]["first_residue_number"] = first_resn
        assert int(self.metadata["protein"]["first_residue_number"]) == first_resn

        # Split variant in amino acid letters and residue numbers
        aa_ref_ll = list(self.dataframe["aa_ref"])
        resi_ll = list(self.dataframe["resi"])
        aa_mut_ll = list(self.dataframe["aa_var"])

        # Update all variants
        new_var_list = []
        ivar_rm = []
        for ivar in range(len(resi_ll)):
            subst_list = []
            for isub in range(len(resi_ll[ivar])):
                # 0-based index of original sequence without offset
                resi_orig = resi_ll[ivar][isub] - first_resn_meta
                # Is substitution marked for removal?
                if resi_orig in resi_rm:
                    ivar_rm.append(ivar)
                    break
                # Update residue number
                resi_ll[ivar][isub] += resi_shift[resi_orig]
                # Update reference amino acid
                if resi_orig in aa_change.keys():
                    if aa_change[resi_orig] == aa_mut_ll[ivar][isub]:
                        if verbose > 2:
                            print(
                                "Data contains a substitution, %s%d%s, into the new reference which will be removed"
                                % (
                                    aa_ref_ll[ivar][isub],
                                    resi_ll[ivar][isub],
                                    aa_mut_ll[ivar][isub],
                                )
                            )
                        ivar_rm.append(ivar)
                        break
                    else:
                        if verbose > 2:
                            print(
                                "Convert variant due to mismatch: %s%d%s -> %s%d%s"
                                % (
                                    aa_ref_ll[ivar][isub],
                                    resi_ll[ivar][isub] - resi_shift[resi_orig],
                                    aa_mut_ll[ivar][isub],
                                    aa_change[resi_orig],
                                    resi_ll[ivar][isub],
                                    aa_mut_ll[ivar][isub],
                                )
                            )
                        aa_ref_ll[ivar][isub] = aa_change[resi_orig]
                subst_list.append(
                    "%s%d%s"
                    % (
                        aa_ref_ll[ivar][isub],
                        resi_ll[ivar][isub],
                        aa_mut_ll[ivar][isub],
                    )
                )
            if len(subst_list) > 0:
                new_var_list.append(":".join(subst_list))
            else:
                new_var_list.append("WT")

        if verbose > 0 and len(ivar_rm) > 0:
            rm_list = [str(self.dataframe["variant"].iat[i]) for i in ivar_rm]
            if len(rm_list) > 500:
                print(
                    "New reference sequence will remove %d variants (mismatch setting: %s; indels: %d):\n%s\n... [skip %d variants] ...\n%s"
                    % (
                        len(rm_list),
                        mismatch,
                        n_indel,
                        ",".join(rm_list[0:25]),
                        len(rm_list) - 50,
                        ",".join(rm_list[-25:]),
                    )
                )
            else:
                print(
                    "New reference sequence will remove %d variants (mismatch setting: %s; indels: %d):\n%s"
                    % (len(rm_list), mismatch, n_indel, ",".join(rm_list))
                )

        self.dataframe["variant"] = new_var_list
        self.dataframe["aa_ref"] = aa_ref_ll
        self.dataframe["resi"] = resi_ll
        self.dataframe["aa_var"] = aa_mut_ll
        # Remove rows containing variants with deleted substitutions
        self.dataframe = self.dataframe.drop(self.dataframe.index[ivar_rm], axis=0)
        self.dataframe.index = range(self.dataframe.shape[0])

    def merge(
        self, data_list, target_seq=None, first_resn=None, merge="outer", **kwargs
    ):
        """Merge a list of data files into a single DataFrame and dump it in a file

        Parameters
        ----------
        data_list : list of VariantData
            The list of data sets to merge
        target_seq : string
            Amino acid sequence that all data sets are aligned to. If None,
            the sequence of the first data set will be used.
        first_resn : Integer >= 0
            Residue number of first amino acid in target_seq, default one. If
            target_seq is None, the header value will be used if present or
            overwritten if an argument is given here.
        merge : string
            How to merge:
            - outer: output the union of variants, default
            - left: Only keep veriants that are present in the self data set
            - inner: output the intersect of variants
        **kwargs : keyword arguments
            Passed to to_new_reference function
        """
        if merge not in ["left", "outer", "inner"]:
            raise ValueError("Allowed merge arguments are left, outer or inner")

        merged_data = self.copy()

        if isinstance(data_list, VariantData):
            data_list = [data_list]
        elif len(data_list) < 1:
            raise ValueError("List of data set to merge is empty")

        # Target sequence
        if target_seq is None:
            # Make from meta data, variant residue numbers will match the index of this
            target_seq = self.metadata["protein"]["sequence"]
            if first_resn is not None:
                raise ValueError(
                    "merge argument first_resn can only be set if target_seq != None\n"
                    + "Use VariantData.to_new_reference to only shift residue numbering"
                )
            first_resn = (
                int(self.metadata["protein"]["first_residue_number"])
                if "first_residue_number" in self.metadata["protein"].keys()
                else 1
            )
        else:
            # Align self data set to target
            if first_resn is None:
                first_resn = 1
            else:
                if first_resn < 0:
                    raise ValueError("Argument first_resn must be non-negative")
            try:
                merged_data.to_new_reference(target_seq, first_resn, **kwargs)
            except PrismValueFail as msg:
                print("Could not align data to target sequence: %s" % (msg))
                return None

        # Remove the index columns ['n_mut','aa_ref','resi','aa_mut'] before merging
        merged_data.remove_index_columns()
        # Add '_00' to the names of all but variant column
        merged_data.dataframe.columns = [merged_data.dataframe.columns[0]] + [
            s + "_00" for s in merged_data.dataframe.columns[1:]
        ]

        # Init new meta data dictionary
        merged_metadata = {
            "version": 0,
            "protein": {},
            "merged": {},
            "variants": {},
            "columns": {},
        }
        merged_metadata["protein"] = copy.deepcopy(self.metadata["protein"])

        merged_metadata["merged"]["file_00"] = (
            self.metadata["filename"]
            + " (version "
            + str(self.metadata["version"])
            + ")"
        )
        for key in self.metadata["columns"].keys():
            merged_metadata["columns"][key + "_00"] = self.metadata["columns"][key]

        # Align and merge other data sets
        c = 0
        for data in data_list:
            data_copy = data.copy()
            # 2DO: Consider to check uniprot (others are human readable and may have typoes) of data to be merged
            try:
                data_copy.to_new_reference(target_seq, first_resn, **kwargs)
            except PrismValueFail as msg:
                print(
                    "Skip merge of file %s: %s" % (data_copy.metadata["filename"], msg)
                )
                continue
            data_copy.remove_index_columns()
            suffix = "_%02d" % (c + 1)
            data_copy.dataframe.columns = [data_copy.dataframe.columns[0]] + [
                s + suffix for s in data_copy.dataframe.columns[1:]
            ]
            merged_data.dataframe = merged_data.dataframe.merge(
                data_copy.dataframe,
                on="variant",
                how=merge,
                suffixes=(False, False),
                copy=False,
            )
            if (
                merged_metadata["protein"]["uniprot"]
                != data.metadata["protein"]["uniprot"]
            ):
                print(
                    "WARNING: Uniprot mismatch on merge: Merging file %s uniprot %s with file %s of different uniprot %s"
                    % (
                        merged_metadata["merged"]["file_00"],
                        merged_metadata["protein"]["uniprot"],
                        data.metadata["filename"],
                        data.metadata["protein"]["uniprot"],
                    )
                )
            merged_metadata["merged"]["file_%02d" % (c + 1)] = (
                data_copy.metadata["filename"]
                + " (version "
                + str(data_copy.metadata["version"])
                + ")"
            )
            for key in data_copy.metadata["columns"].keys():
                merged_metadata["columns"][key + "_%02d" % (c + 1)] = (
                    data_copy.metadata["columns"][key]
                )
            c += 1

        # Add the index columns ['n_mut','aa_ref','resi','aa_mut'] again
        merged_data.add_index_columns()

        merged_metadata["protein"]["sequence"] = target_seq
        if first_resn != 1:
            merged_metadata["protein"]["first_residue_number"] = first_resn
        merged_data.metadata = merged_metadata
        merged_data.metadata["variants"] = merged_data.calc_variants_metadata()
        return merged_data


# class ResidueData(PrismData):
#     """Container class for per-residue data
#     """
#
#     def _init(self):
#         self.index_column_names = ['aa_ref']
#
#     def add_index_columns(self):
#         pass
#
#     def remove_index_columns(self):
#         """Remove the indexing columns, i.e. 'aa_ref'"""
#         self.dataframe.drop(['aa_ref'], axis=1, inplace=True)
#
#     def check(self, verbose=0):
#         # check ResidueData specific stuff
#         check_status = None
#         try:
#             self.check_column_names(verbose=verbose)
#             self.check_header_and_data_sequence(verbose=verbose)
#             self.check_variants_metadata(no_overwrite=no_overwrite_metadata, verbose=verbose)
#         except PrismFormatError as msg:
#             raise PrismFormatError("VariantData.check failed. First fail message:\n%s" % (msg))
#         else:
#             check_status = True
#
#         return(check_status)
#
#     def to_new_reference(self, target_seq=None, first_resn=None, verbose=0):
#         pass


if __name__ == "__main__":
    # Parse commandline arguments
    import argparse
    import sys

    arg_parser = argparse.ArgumentParser(
        description="PRISM data file processing and alignment"
    )
    # positional arguments
    arg_parser.add_argument(
        "files", nargs="+", metavar="DATAFILE", help="Data files to process"
    )
    # keyword arguments
    arg_parser.add_argument(
        "--dump_header_csv",
        metavar="FILE",
        help="Dump a .csv file containing an overview of file headers",
    )
    arg_parser.add_argument(
        "--dump_fasta", metavar="FILE", help="Dump all sequences in a fasta file"
    )
    arg_parser.add_argument(
        "--parse",
        action="store_true",
        help="Read data file(s), run all checks and write new data file(s)",
    )
    arg_parser.add_argument(
        "--merge",
        metavar="FILE",
        help="Merge data files concerning the (approximately) same target protein",
    )
    # A run of this main can only have a single target sequence so to parse more files with different sequence, run multiple times
    arg_parser.add_argument(
        "--target_seq",
        metavar="FILE_OR_SEQUENCE",
        help="Align all output to this sequence, default is the sequence given in the first file",
    )
    arg_parser.add_argument(
        "--saturate",
        action="store_true",
        help="In the output data, add rows to cover all 20 substitutions at all considered positions",
    )
    arg_parser.add_argument(
        "--clean_na",
        action="store_true",
        help="In the output data, add rows to cover all 20 substitutions at all considered positions",
    )
    arg_parser.add_argument(
        "--first_res_num",
        metavar="INT",
        help="In the output data, assign this number to the first given amino acid of the sequence",
    )
    arg_parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Level of output, default zero is no output",
    )
    args = arg_parser.parse_args()
    if args.saturate and args.clean_na:
        raise ValueError("Arguments saturate and clean_na cannot be used together")

    # Read data files
    data_parser = PrismParser()
    data_list = []
    for filename in args.files:
        try:
            data = None
            if args.merge or args.parse:
                if args.verbose > 0:
                    print("\nRead data file %s " % (filename), flush=True)
                data = data_parser.read(filename, verbose=args.verbose)
            elif args.dump_header_csv or args.dump_fasta:
                # Only read header which is much faster for large files
                if args.verbose > 0:
                    print("\nRead header of %s " % (filename), flush=True)
                header = data_parser.read_header(filename, verbose=args.verbose)
                data = PrismData(header, None)
            else:
                break
        except PrismFormatError as error:
            print("ERROR reading file %s:\nBad file format: %s\n" % (filename, error))
        except yaml.parser.ParserError as error:
            print(
                "ERROR reading file %s:\nBad YAML syntax (ParserError) in header:\n%s\n"
                % (filename, error)
            )
        except yaml.scanner.ScannerError as error:
            print(
                "ERROR reading file %s:\nBad YAML syntax (ScannerError, perhaps a missing whitespace after colon?) in header:\n%s\n"
                % (filename, error)
            )
        except FileNotFoundError:
            print("ERROR reading file %s:\nCannot find file\n" % (filename))
        else:
            # Only if all is good
            data_list.append(data)

    # Check input
    if len(data_list) < 1:
        print("\nERROR: Missing data files or arguments")
        arg_parser.print_usage()
        sys.exit(2)
    if args.verbose > 0:
        print("Done loading %d file(s)\n" % (len(data_list)))

    # Dump overview csv if requested
    if args.dump_header_csv:
        if args.target_seq or args.first_res_num or args.saturate or args.clean_na:
            print(
                "WARNING: Arguments target_seq, first_res_num, saturate and clean_na have no effect on dump_header_csv"
            )
        data_parser.write_meta_file(data_list, args.dump_header_csv, output_type="csv")
        print(
            "Dumped header overveiw of %d files in %s"
            % (len(data_list), args.dump_header_csv)
        )

    # Dump amino acid sequences in fasta format if requested
    if args.dump_fasta:
        if args.target_seq or args.first_res_num or args.saturate or args.clean_na:
            print(
                "WARNING: Arguments target_seq, first_res_num, saturate and clean_na have no effect on dump_fasta"
            )
        data_parser.write_meta_file(data_list, args.dump_fasta, output_type="fasta")
        print("Dumped %d sequences in %s" % (len(data_list), args.dump_fasta))

    # Set target sequence and offset for all data files
    target_seq = None
    first_resn = 1
    if args.target_seq:
        if (
            args.target_seq[-6:].lower() == ".fasta"
            or args.target_seq[-4:].lower() == ".fas"
        ):
            record = None
            for r in SeqIO.parse(args.target_seq, "fasta"):
                if record is not None:
                    # if args.verbose > 0:
                    print(
                        "WARNING: Only using the first sequence record in %s"
                        % (args.target_seq)
                    )
                    break
                record = r
            target_seq = str(record.seq)
            print(
                "Target sequence '%s' of %d residues from file %s"
                % (record.description, len(target_seq), args.target_seq)
            )
        else:
            target_seq = args.target_seq
        if args.first_res_num:
            first_resn = int(args.first_res_num)
    else:
        target_seq = (data_list[0]).metadata["protein"]["sequence"]
        if args.first_res_num:
            # Overwrite offset given in header of first file
            first_resn = int(args.first_res_num)
        elif "first_residue_number" in (data_list[0]).metadata["protein"]:
            # Only use offset of first file with that sequence (and not the commandline input)
            first_resn = int((data_list[0]).metadata["protein"]["first_residue_number"])

    if not data_parser.is_aa_one_nat(target_seq, "X"):
        print("ERROR: Target sequence seems not to be one-letter amino acids or 'X':")
        print("Target sequence: %s" % (target_seq))
        sys.exit(2)

    # Parse & dump all files if requested
    if args.parse:
        filename_suffix = "_parsed"
        for data in data_list:
            if args.target_seq or first_resn:
                data.to_new_reference(
                    target_seq,
                    first_resn,
                    verbose=args.verbose,
                    min_identity=0.9,
                    min_coverage=0.1,
                    mismatch="remove",
                    allow_inserts=True,
                    allow_deletions=True,
                )
            if args.saturate:
                data.saturate_variants()
            elif args.clean_na:
                data.clean_na_variants()
            filename_parts = data.metadata["filename"].split("/")[-1].rsplit(".", 1)
            outfilename = filename_parts[0] + filename_suffix + "." + filename_parts[1]
            data_parser.write(outfilename, data)
        print(
            "Parsed and wrote %d data file(s) with output filename suffix %s"
            % (len(data_list), filename_suffix)
        )

    # Dump a merged data file if requested
    if args.merge:
        merged_data = data_list[0].merge(
            data_list[1:],
            target_seq,
            first_resn,
            merge="outer",
            verbose=args.verbose,
            min_identity=0.9,
            min_coverage=0.1,
            mismatch="remove",
            allow_inserts=True,
            allow_deletions=True,
        )
        if merged_data is None:
            print("ERROR: Could not merge data")
            sys.exit(2)

        if args.saturate:
            merged_data.saturate_variants()
        elif args.clean_na:
            merged_data.clean_na_variants()
        data_parser.write(args.merge, merged_data)
        print(
            "Dumped merge of %d data files in %s "
            % (len(merged_data.metadata["merged"]), args.merge)
        )



from poli.core.black_box_information import BlackBoxInformation
from poli.core.util.proteins.defaults import AMINO_ACIDS

gfp_select_info = BlackBoxInformation(
    name="gfp_select",
    max_sequence_length=237,  # max len of aaSequence
    aligned=True,  # TODO: perhaps add the fact that there is a random state here?
    fixed_length=True,
    deterministic=False,
    alphabet=AMINO_ACIDS,
    log_transform_recommended=False,
    discrete=True,
    fidelity=None,
    padding_token="",
)



"""
Implementation of the GuacaMol benchmark.

GuacaMol [1] is a benchmark for small molecule optimization. It was
developed in the context of generative models for molecular design.
Originally, it had two components: distirbution-learning and goal-directed
benchmarks. This implementation focuses on goal-directed benchmarks.

Our implementation heavily relies on Therapeutics Data Commons [2].

Check the details in the original paper [1].

References
----------
[1] Brown, Nathan, Marco Fiscato, Marwin H.S. Segler, and Alain C. Vaucher.
    “GuacaMol: Benchmarking Models for de Novo Molecular Design.”
    Journal of Chemical Information and Modeling 59, no. 3 (March 25, 2019):
    1096-1108. https://doi.org/10.1021/acs.jcim.8b00839.

[2] Huang, Kexin, Tianfan Fu, Wenhao Gao, Yue Zhao, Yusuf Roohani, Jure Leskovec,
    Connor W Coley, Cao Xiao, Jimeng Sun, and Marinka Zitnik.
    “Therapeutics Data Commons: Machine Learning Datasets and Tasks for Drug Discovery and Development.”
    Proceedings of Neural Information Processing Systems, NeurIPS Datasets and Benchmarks, 2021.
"""

from typing import Literal, Union

from poli.core.abstract_benchmark import AbstractBenchmark
from poli.core.problem import Problem
from poli.objective_repository import (
    AlbuterolSimilarityProblemFactory,
    AmlodipineMPOProblemFactory,
    CelecoxibRediscoveryProblemFactory,
    DecoHopProblemFactory,
    FexofenadineMPOProblemFactory,
    IsomerC7H8N2O2ProblemFactory,
    IsomerC9H10N2O2PF2ClProblemFactory,
    LogPProblemFactory,
    Median1ProblemFactory,
    Median2ProblemFactory,
    MestranolSimilarityProblemFactory,
    OsimetrinibMPOProblemFactory,
    PerindoprilMPOProblemFactory,
    QEDProblemFactory,
    RanolazineMPOProblemFactory,
    SAProblemFactory,
    ScaffoldHopProblemFactory,
    SitagliptinMPOProblemFactory,
    ThiothixeneRediscoveryProblemFactory,
    TroglitazoneRediscoveryProblemFactory,
    ValsartanSMARTSProblemFactory,
    ZaleplonMPOProblemFactory,
)


class GuacaMolGoalDirectedBenchmark(AbstractBenchmark):
    """GuacaMol benchmark for goal-directed tasks

    GuacaMol [1] is a benchmark for small molecule optimization. It was
    developed in the context of generative models for molecular design.
    Originally, it had two components: distirbution-learning and goal-directed
    benchmarks. This implementation focuses on goal-directed benchmarks.

    Our implementation heavily relies on Therapeutics Data Commons [2].

    Check the details in the original paper [1].

    References
    ----------
    [1] Brown, Nathan, Marco Fiscato, Marwin H.S. Segler, and Alain C. Vaucher.
        “GuacaMol: Benchmarking Models for de Novo Molecular Design.”
        Journal of Chemical Information and Modeling 59, no. 3 (March 25, 2019):
        1096-1108. https://doi.org/10.1021/acs.jcim.8b00839.

    [2] Huang, Kexin, Tianfan Fu, Wenhao Gao, Yue Zhao, Yusuf Roohani, Jure Leskovec,
        Connor W Coley, Cao Xiao, Jimeng Sun, and Marinka Zitnik.
        “Therapeutics Data Commons: Machine Learning Datasets and Tasks for Drug Discovery and Development.”
        Proceedings of Neural Information Processing Systems, NeurIPS Datasets and Benchmarks, 2021.
    """

    def __init__(
        self,
        string_representation: Literal["SMILES", "SELFIES"],
        seed: Union[int, None] = None,
        batch_size: Union[int, None] = None,
        parallelize: bool = False,
        num_workers: Union[int, None] = None,
        evaluation_budget: int = float("inf"),
    ) -> None:
        super().__init__(
            seed=seed,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )

        self.problem_factories = [
            AlbuterolSimilarityProblemFactory(),
            AmlodipineMPOProblemFactory(),
            CelecoxibRediscoveryProblemFactory(),
            DecoHopProblemFactory(),
            FexofenadineMPOProblemFactory(),
            IsomerC7H8N2O2ProblemFactory(),
            IsomerC9H10N2O2PF2ClProblemFactory(),
            Median1ProblemFactory(),
            Median2ProblemFactory(),
            MestranolSimilarityProblemFactory(),
            OsimetrinibMPOProblemFactory(),
            PerindoprilMPOProblemFactory(),
            RanolazineMPOProblemFactory(),
            LogPProblemFactory(),
            QEDProblemFactory(),
            SAProblemFactory(),
            ScaffoldHopProblemFactory(),
            SitagliptinMPOProblemFactory(),
            ThiothixeneRediscoveryProblemFactory(),
            TroglitazoneRediscoveryProblemFactory(),
            ValsartanSMARTSProblemFactory(),
            ZaleplonMPOProblemFactory(),
        ]
        self.string_representation = string_representation

    def _initialize_problem(self, index: int) -> Problem:
        problem_factory = self.problem_factories[index]

        problem = problem_factory.create(
            string_representation=self.string_representation,
            seed=self.seed,
            batch_size=self.batch_size,
            parallelize=self.parallelize,
            num_workers=self.num_workers,
            evaluation_budget=self.evaluation_budget,
        )

        return problem



"""
Implements a wrapper class for the CBas VAE loading implementation.
Originally implemented in SB's discrete-bo repository.
Only minor modifications conducted by RM.

Creator: SB
Last Changes: RM
"""

import torch


class AbstractVAEWrapper:
    def encode_sequences(self, seqs: [torch.Tensor]) -> torch.Tensor:
        raise NotImplementedError("abstract method")

    def encode(self, x: torch.Tensor) -> torch.Tensor:
        """
        Turns a one hot encoded LxAA vector into a continuous lower dimensional vector.
        :param x:
        :type x:
        :return:
        :rtype:
        """
        raise NotImplementedError("abstract method")

    def decode(self, z: torch.Tensor, grad=False) -> torch.Tensor:
        """
        Returns an NxLxAA tensor.
        :param z:
        :type z:
        :param grad:
        :type grad:
        :return:
        :rtype:
        """
        raise NotImplementedError("abstract method")

    def transform_sequences_to_atoms(self, x: [torch.Tensor]) -> torch.Tensor:
        """
        Turns a sequence into a Dirac distribution around the sequence.
        NO ZERO-HANDLING!
        :param x:
            a list of sequences encoded according to the problem alphabet
        :type x:
        :return:
        Returns an NxLxAA tensor.
        :rtype:
        """
        raise NotImplementedError("abstract method")

    def decode_samples_from_encoder(
        self, x: torch.Tensor, num_samples: int
    ) -> torch.Tensor:
        """
        For a single one hot encoded LxAA vector this method samples around the encoded sequence and returns the decoded sequences.
        """
        raise NotImplementedError("abstract method")

    def get_latent_dimensionality(self) -> int:
        raise NotImplementedError("abstract method")

    def get_max_llh_sequence(self, proba: torch.Tensor):
        raise NotImplementedError("abstract method")

    def get_wildtype_distribution(self):
        """Returns a distribution to represent the wild-type."""
        raise NotImplementedError("abstract method")



"""
Implements the Isomer C7H8N2O2 task using the TDC oracles [1].

This task is inherited from the GuacaMol benchmark [2]. We recommend
you cite both references when using this task.

References
----------
[1] Artificial intelligence foundation for therapeutic science.
    Huang, K., Fu, T., Gao, W. et al.  Nat Chem Biol 18, 1033-1036 (2022). https://doi.org/10.1038/s41589-022-01131-2
[2] GuacaMol: benchmarking models for de novo molecular design.
    Brown, N. et al.  J Chem Inf Model 59 (2019).
    https://pubs.acs.org/doi/10.1021/acs.jcim.8b00839
"""

from __future__ import annotations

from typing import Literal

import numpy as np
import selfies as sf

from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.chemistry.tdc_black_box import TDCBlackBox
from poli.core.chemistry.tdc_problem import TDCProblem
from poli.core.problem import Problem
from poli.core.util.chemistry.string_to_molecule import translate_smiles_to_selfies
from poli.core.util.seeding import seed_numpy, seed_python


class IsomerC7H8N2O2BlackBox(TDCBlackBox):
    """
    A black box that measures the similarities of a certain
    SMILES string's Isomer presentation to C7H8N2O2, implemented
    using the TDC oracles [1].

    This task is inherited from the GuacaMol benchmark [2]. We
    recommend you cite both references when using this task.

    Parameters
    ----------
    string_representation : Literal["SMILES", "SELFIES"], optional
        A string (either "SMILES" or "SELFIES") specifying which
        molecule representation you plan to use.
    alphabet : list[str] | None, optional
        The alphabet to be used for the SMILES or SELFIES representation.
        It is common that the alphabet depends on the dataset used, so
        it is recommended to pass it as an argument. Default is None.
    max_sequence_length : int, optional
        The maximum length of the sequence. Default is infinity.
    batch_size : int, optional
        The batch size for simultaneous execution, by default None.
    parallelize : bool, optional
        Flag indicating whether to parallelize execution, by default False.
    num_workers : int, optional
        The number of workers for parallel execution, by default None.
    evaluation_budget:  int, optional
        The maximum number of function evaluations. Default is infinity.
    force_isolation: bool, optional
        Whether to force the isolation of the black box. Default is False.

    Attributes
    ----------
    oracle_name : str
        The name of the oracle.

    Methods
    -------
    __init__(self, string_representation, force_isolation=False, batch_size=None, parallelize=False, num_workers=None, evaluation_budget=float('inf'))
        Initializes a new instance of the black box.

    References
    ----------
    [1] Artificial intelligence foundation for therapeutic science.
        Huang, K., Fu, T., Gao, W. et al.  Nat Chem Biol 18, 1033-1036 (2022). https://doi.org/10.1038/s41589-022-01131-2
    [2] GuacaMol: benchmarking models for de novo molecular design.
        Brown, N. et al.  J Chem Inf Model 59 (2019).
        https://pubs.acs.org/doi/10.1021/acs.jcim.8b00839
    """

    def __init__(
        self,
        string_representation: Literal["SMILES", "SELFIES"] = "SMILES",
        alphabet: list[str] | None = None,
        max_sequence_length: int = np.inf,
        force_isolation: bool = False,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
    ):
        super().__init__(
            oracle_name="Isomers_C7H8N2O2",
            string_representation=string_representation,
            alphabet=alphabet,
            max_sequence_length=max_sequence_length,
            force_isolation=force_isolation,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )

    def get_black_box_info(self) -> BlackBoxInformation:
        return BlackBoxInformation(
            name="isomer_c7h8n2o2",
            max_sequence_length=self.max_sequence_length,
            aligned=False,
            fixed_length=False,
            deterministic=True,  # ?
            alphabet=self.alphabet,  # TODO: add alphabet once we settle for one for SMLIES/SELFIES.
            log_transform_recommended=False,
            discrete=True,
            padding_token="",
        )


class IsomerC7H8N2O2ProblemFactory(AbstractProblemFactory):
    """
    Factory class for creating Isomer C7H8N2O2 problems.

    This task is inherited from the GuacaMol benchmark [2]. We
    recommend you cite [1, 2] when using this task.

    Methods
    ------
    create:
        Creates an Isomer C7H8N2O2 problem, containing a black box
        and an initial value x0 (taken from the documentation of TDC).

    References
    ----------
    [1] Artificial intelligence foundation for therapeutic science.
        Huang, K., Fu, T., Gao, W. et al.  Nat Chem Biol 18, 1033-1036 (2022). https://doi.org/10.1038/s41589-022-01131-2
    [2] GuacaMol: benchmarking models for de novo molecular design.
        Brown, N. et al.  J Chem Inf Model 59 (2019).
        https://pubs.acs.org/doi/10.1021/acs.jcim.8b00839
    """

    def create(
        self,
        string_representation: Literal["SMILES", "SELFIES"] = "SMILES",
        alphabet: list[str] | None = None,
        max_sequence_length: int = np.inf,
        seed: int = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        """
        Creates an Isomer C7H8N2O2 problem.

        Parameters
        ----------
        string_representation : str, optional
            The string representation of the molecules. Must be either 'SMILES' or 'SELFIES'. Default is 'SMILES'.
        alphabet : list[str] | None, optional
            The alphabet to be used for the SMILES or SELFIES representation.
            It is common that the alphabet depends on the dataset used, so
            it is recommended to pass it as an argument. Default is None.
        max_sequence_length : int, optional
            The maximum length of the sequence. Default is infinity.
        seed : int, optional
            Seed for random number generators. If None, no seed is set.
        batch_size : int, optional
            Number of molecules to process in parallel. If None, the default batch size is used.
        parallelize : bool, optional
            Whether to parallelize the docking process. Default is False.
        num_workers : int, optional
            Number of worker processes to use for parallelization. If None, the number of available CPU cores is used.
        evaluation_budget:  int, optional
            The maximum number of function evaluations. Default is infinity.
        force_isolation: bool, optional
            Whether to force the isolation of the black box. Default is False.

        Returns
        -------
        problem : Problem
            A problem instance containing the black box, and an initial value x0.

        Raises
        ------
        ValueError
            If the string_representation is not 'SMILES' or 'SELFIES'.
        """
        # We start by seeding the RNGs
        if seed is not None:
            seed_numpy(seed)
            seed_python(seed)

        # We check whether the string representation is valid
        if string_representation.upper() not in ["SMILES", "SELFIES"]:
            raise ValueError(
                "Missing required keyword argument: string_representation: str. "
                "String representation must be either 'SMILES' or 'SELFIES'."
            )

        f = IsomerC7H8N2O2BlackBox(
            string_representation=string_representation,
            alphabet=alphabet,
            max_sequence_length=max_sequence_length,
            force_isolation=force_isolation,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )

        # Initial example (from the TDC docs)
        x0_smiles = "CC(C)(C)[C@H]1CCc2c(sc(NC(=O)COc3ccc(Cl)cc3)c2C(N)=O)C1"
        x0_selfies = translate_smiles_to_selfies([x0_smiles])[0]

        if string_representation.upper() == "SMILES":
            x0 = np.array([list(x0_smiles)])
        else:
            x0 = np.array([list(sf.split_selfies(x0_selfies))])

        isomer_problem = TDCProblem(
            black_box=f,
            x0=x0,
        )

        return isomer_problem



"""A closed-form black box simulating epistatic effects [1].

References
----------
[1] Stanton, S., Alberstein, R., Frey, N., Watkins, A., & Cho, K. (2024).
    Closed-Form Test Functions for Biophysical Sequence Optimization Algorithms.
    arXiv preprint arXiv:2407.00236. https://arxiv.org/abs/2407.00236
"""

from .register import EhrlichHoloBlackBox, EhrlichHoloProblemFactory

__all__ = ["EhrlichHoloBlackBox", "EhrlichHoloProblemFactory"]



"""This module contains utilities for creating run scripts for problems and observers.
"""

import inspect
import os
import stat
import sys
from os.path import basename, dirname, join
from pathlib import Path
from typing import List, Type, Union

from poli import external_isolated_function_script
from poli.core.abstract_isolated_function import AbstractIsolatedFunction
from poli.core.util import observer_wrapper
from poli.core.util.abstract_observer import AbstractObserver
from poli.external_isolated_function_script import ADDITIONAL_IMPORT_SEARCH_PATHES_KEY

# By default, we will store the run scripts inside the
# home folder of the user, on the hidden folder
# ~/.poli_objectives
HOME_DIR = Path.home().resolve()
RUN_SCRIPTS_FOLDER = HOME_DIR / ".poli_objectives"


def make_isolated_function_script(
    isolated_function: AbstractIsolatedFunction,
    conda_environment_name: Union[str, Path] = None,
    python_paths: List[str] = None,
    cwd=None,
    **kwargs,
):
    """
    Create a script to run the given black box, returning its location.

    Parameters
    ----------
    black_box : AbstractBlackBox
        The black box object to be executed.
    conda_environment_name : str or Path, optional
        The conda environment to activate before running the black box.
    python_paths : List[str], optional
        Additional Python paths to be added before running the black box.
    cwd : str or Path, optional
        The current working directory for the script execution.

    Returns
    -------
    run_script: str
        The path to the generated script.

    """
    command = inspect.getfile(external_isolated_function_script)
    return _make_run_script_from_template(
        command, isolated_function, conda_environment_name, python_paths, cwd, **kwargs
    )


def make_observer_script(
    observer: Type[AbstractObserver],
    conda_environment: Union[str, Path] = None,
    python_paths: List[str] = None,
    cwd=None,
):
    """
    Create a script to run the given observer.

    Parameters
    ----------
    observer : AbstractObserver
        The observer object to be executed.
    conda_environment : str or Path, optional
        The conda environment to activate before running the observer.
    python_paths : List[str], optional
        Additional Python paths to be added before running the observer.
    cwd : str or Path, optional
        The current working directory for the script execution.

    Returns
    -------
    run_script: str
        The path to the generated script.

    """
    command = inspect.getfile(observer_wrapper)
    return _make_run_script_from_template(
        command, observer, conda_environment, python_paths, cwd
    )


def _make_run_script_from_template(
    command: str,
    non_instantiated_object,
    conda_environment_name: Union[str, Path],
    python_paths: List[str],
    cwd=None,
):
    """
    An internal function for creating run scripts; returns the location of the run script.

    Parameters
    ----------
    command : str
        The command to be executed in the run script.
    instantiated_object : object
        The instantiated object representing the problem factory.
    conda_environment_name : str or Path
        The name or path of the conda environment to be used.
    python_paths : List[str]
        The list of python paths to be appended to the run script.
    cwd : str, optional
        The current working directory for the run script. If not provided, the current working directory is used.
    **kwargs : dict
        Additional keyword arguments (currently unused).

    Returns
    -------
    run_script: str
        The location of the generated run script.
    """
    if cwd is None:
        cwd = str(os.getcwd())

    # class_object = instantiated_object.__class__
    class_object = non_instantiated_object
    problem_factory_name = class_object.__name__  # TODO: potential vulnerability?
    factory_location = inspect.getfile(class_object)
    package_name = inspect.getmodule(non_instantiated_object).__name__

    if package_name == "__main__":
        package_name = basename(factory_location)[:-3]

    # else:
    package_name += "."

    full_problem_factory_name = package_name + problem_factory_name
    run_script_location = join(RUN_SCRIPTS_FOLDER, problem_factory_name + ".sh")

    if isinstance(conda_environment_name, str):
        # TODO: check that conda environment exists
        ...
    elif isinstance(conda_environment_name, Path):
        conda_environment_name = conda_environment_name.resolve()
        if not conda_environment_name.exists():
            raise ValueError(
                f"conda_environment_location {conda_environment_name} does not exist."
            )

        conda_environment_name = str(conda_environment_name)
    else:
        if conda_environment_name is None:
            conda_environment_name = sys.executable[: -len("/bin/python")]
            conda_environment_name = str(os.path.abspath(conda_environment_name))
        else:
            raise ValueError(
                "If specified, conda_environment_location must be a string or a Path."
            )

    # make path to conda environment absolute
    if python_paths is None:
        python_paths = [dirname(factory_location)]

    # TODO: check that location exists and is valid environment
    python_paths = ":".join(python_paths)

    with open(
        join(dirname(__file__), "run_script_template.sht"), "r"
    ) as run_script_template_file:
        """
        load template file and fill in gaps which load conda enviroment, append python paths and call run with the
        desired factory
        """
        run_script = run_script_template_file.read() % (
            cwd,
            conda_environment_name,
            python_paths,
            ADDITIONAL_IMPORT_SEARCH_PATHES_KEY,
            command,
            full_problem_factory_name,
        )
    with open(run_script_location, "w+") as run_script_file:
        # write out run script and make it executable
        run_script_file.write(run_script)
        os.chmod(
            run_script_location, os.stat(run_script_location).st_mode | stat.S_IEXEC
        )  # make script file executable

    return run_script_location
